import copy
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, cast

import requests
from requests import RequestException
from typing_extensions import NotRequired, TypedDict

from gpuhunt._internal.constraints import find_accelerators
from gpuhunt._internal.models import AcceleratorVendor, QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
API_URL = "https://api.runpod.io/graphql"


class RunpodCatalogItemProviderData(TypedDict):
    # `pod_counts` is the number of pods that can be deployed in a cluster for the given offer.
    # Used to distinguish Runpod Clusters offers and check if multinode runs fit into clusters.
    pod_counts: NotRequired[list[int]]


class RunpodProvider(AbstractProvider):
    NAME = "runpod"
    # Minimum CUDA version on the host. Used to filter available offers
    # and should also be used when provisioning pods.
    MIN_CUDA_VERSION = "12.8"

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        offers = self.fetch_offers()
        return sorted(offers, key=lambda i: i.price)

    @staticmethod
    def fetch_offers() -> list[RawCatalogItem]:
        query_variables = build_query_variables()
        with ThreadPoolExecutor(max_workers=10) as executor:
            futures = [
                executor.submit(get_pods, query_variable) for query_variable in query_variables
            ]
        pods_by_query = []
        for future in futures:
            try:
                pods_by_query.append(future.result())
            except RequestException as e:
                logger.exception("Failed to get pods data: %s", e)

        catalog_items = []
        for query_variable, pods in zip(query_variables, pods_by_query):
            for pod in pods:
                catalog_items.extend(make_catalog_items(query_variable, pod))

        cluster_catalog_items = fetch_cluster_offers()
        catalog_items.extend(cluster_catalog_items)
        return catalog_items

    @classmethod
    def filter(cls, offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
        return [
            o
            for o in offers
            if o.location
            not in [
                "AR",  # network problems, unusable
            ]
        ]


def gpu_vendor_and_name(gpu_id: str) -> Optional[tuple[AcceleratorVendor, str]]:
    if not gpu_id:
        return None
    return GPU_MAP.get(gpu_id)


def build_query_variables() -> list[dict]:
    """Prepare different combinations of API query filters to cover all available GPUs."""

    gpu_types = make_request({"query": gpu_types_query, "variables": {}})
    data_centers = [dc["id"] for dc in gpu_types["data"]["dataCenters"] if dc["listed"]]
    max_gpu_count = max(gpu["maxGpuCount"] for gpu in gpu_types["data"]["gpuTypes"])

    variables = []
    for gpu_count in range(1, max_gpu_count + 1):
        # Secure cloud is queryable by datacenter ID
        for dc_id in data_centers:
            variables.append(
                {
                    "GpuTypeFilter": {
                        "cluster": False,
                    },
                    "lowestPriceInput": {
                        "secureCloud": True,
                        "dataCenterId": dc_id,
                        "gpuCount": gpu_count,
                        "minDisk": None,
                        "minMemoryInGb": None,
                        "minVcpuCount": None,
                        "minCudaVersion": RunpodProvider.MIN_CUDA_VERSION,
                    },
                }
            )
        # Community cloud is queryable by country code
        for country_code in gpu_types["data"]["countryCodes"]:
            if country_code is None:
                continue
            variables.append(
                {
                    "GpuTypeFilter": {
                        "cluster": False,
                    },
                    "lowestPriceInput": {
                        "secureCloud": False,
                        "countryCode": country_code,
                        "gpuCount": gpu_count,
                        "minDisk": None,
                        "minMemoryInGb": None,
                        "minVcpuCount": None,
                        "minCudaVersion": RunpodProvider.MIN_CUDA_VERSION,
                    },
                }
            )
    return variables


def get_pods(query_variables: dict) -> list[dict]:
    resp = make_request(
        {
            "query": query_pod_types,
            "variables": query_variables,
        }
    )
    return resp["data"]["gpuTypes"]


def make_catalog_items(query_variables: dict, pod: dict) -> list[RawCatalogItem]:
    lowest_price_input_variables = query_variables["lowestPriceInput"]
    if pod["lowestPrice"]["stockStatus"] is None:
        return []
    listed_gpu_vendor_and_name = gpu_vendor_and_name(pod["id"])
    if listed_gpu_vendor_and_name is None:
        logger.warning(f"{pod['id']} missing in runpod GPU_MAP")
        return []
    if lowest_price_input_variables["secureCloud"]:
        location = lowest_price_input_variables["dataCenterId"]
        on_demand_gpu_price = pod["securePrice"]
        spot_gpu_price = pod["secureSpotPrice"]
    else:
        location = lowest_price_input_variables["countryCode"]
        on_demand_gpu_price = pod["communityPrice"]
        spot_gpu_price = pod["communitySpotPrice"]
    item_template = RawCatalogItem(
        instance_name=pod["id"],
        location=location,
        price=None,  # set below
        cpu=pod["lowestPrice"]["minVcpu"],
        memory=pod["lowestPrice"]["minMemory"],
        gpu_vendor=listed_gpu_vendor_and_name[0],
        gpu_count=lowest_price_input_variables["gpuCount"],
        gpu_name=listed_gpu_vendor_and_name[1],
        gpu_memory=pod["memoryInGb"],
        spot=None,  # set below
        disk_size=None,
        provider_data={},
    )
    items = []
    if on_demand_gpu_price:
        item = copy.deepcopy(item_template)
        item.spot = False
        item.price = item.gpu_count * on_demand_gpu_price
        items.append(item)
    if spot_gpu_price:
        item = copy.deepcopy(item_template)
        item.spot = True
        item.price = item.gpu_count * spot_gpu_price
        items.append(item)
    return items


def fetch_cluster_offers() -> list[RawCatalogItem]:
    cluster_catalog_items = []
    query_variables = {
        "gpuTypesInput": {
            "cluster": True,
        },
        "lowestPriceInput": {
            "gpuCount": 8,  # Needed to get CPU and RAM
        },
    }
    pod_type = get_pods(query_variables)
    for pod_type in pod_type:
        listed_gpu_vendor_and_name = gpu_vendor_and_name(pod_type["id"])
        if listed_gpu_vendor_and_name is None:
            logger.warning(f"{pod_type['id']} missing in runpod GPU_MAP")
            continue
        gpu_vendor, gpu_name = listed_gpu_vendor_and_name
        # Runpod returns no CPU and memory if the offer is out of stock.
        # Out of stock appears to be different from no capacity meaning
        # the offer is not available for a prolonged period.
        cpu = pod_type["lowestPrice"].get("minVcpu")
        memory = pod_type["lowestPrice"].get("minMemory")
        if cpu is None:
            logger.warning(f"{pod_type['id']} cluster offer missing minVcpu")
            continue
        if memory is None:
            logger.warning(f"{pod_type['id']} cluster offer missing minMemory")
            continue
        for location in pod_type["nodeGroupDatacenters"]:
            catalog_item = RawCatalogItem(
                instance_name=pod_type["id"],
                location=location["id"],
                price=pod_type["clusterPrice"] * pod_type["maxGpuCount"],
                cpu=cpu,
                memory=memory,
                gpu_vendor=gpu_vendor,
                gpu_count=pod_type["maxGpuCount"],
                gpu_name=gpu_name,
                gpu_memory=pod_type["memoryInGb"],
                spot=False,
                disk_size=None,
                flags=["runpod-cluster"],
                # The API does not return supported pod counts but for now it's always 2-8 for all offers.
                provider_data=cast(
                    dict, RunpodCatalogItemProviderData(pod_counts=list(range(2, 9)))
                ),
            )
            cluster_catalog_items.append(catalog_item)
    return cluster_catalog_items


def make_request(payload: dict):
    resp = requests.post(API_URL, json=payload, timeout=10)
    resp.raise_for_status()
    return resp.json()


def get_gpu_map() -> dict[str, tuple[AcceleratorVendor, str]]:
    payload_gpus = {
        "query": "query GpuTypes { gpuTypes { id manufacturer displayName memoryInGb } }"
    }
    response = make_request(payload_gpus)
    gpu_map: dict[str, tuple[AcceleratorVendor, str]] = {}
    for gpu_type in response["data"]["gpuTypes"]:
        try:
            vendor = AcceleratorVendor.cast(gpu_type["manufacturer"])
        except ValueError:
            continue
        gpu_name = get_gpu_name(vendor, gpu_type["displayName"])
        if gpu_name:
            gpu_map[gpu_type["id"]] = (vendor, gpu_name)
    return gpu_map


def get_gpu_name(vendor: AcceleratorVendor, name: str) -> Optional[str]:
    if vendor == AcceleratorVendor.NVIDIA:
        return get_nvidia_gpu_name(name)
    if vendor == AcceleratorVendor.AMD:
        return get_amd_gpu_name(name)
    return None


def get_nvidia_gpu_name(name: str) -> Optional[str]:
    if "B200" in name:
        return "B200"
    if "V100" in name:
        return "V100"
    if name == "H100 NVL":
        return "H100NVL"
    if name.startswith(("A", "L", "H")):
        gpu_name, _, _ = name.partition(" ")
        return gpu_name
    if name.startswith("RTX A"):
        return name.lstrip("RTX ").replace(" ", "")
    if name.startswith("RTX"):
        return name.replace(" ", "")
    return None


def get_amd_gpu_name(name: str) -> Optional[str]:
    if accelerators := find_accelerators(names=[name], vendors=[AcceleratorVendor.AMD]):
        return accelerators[0].name
    return None


GPU_MAP = get_gpu_map()

gpu_types_query = """
query GpuTypes {
  countryCodes
  dataCenters {
    id
    name
    listed
    __typename
  }
  gpuTypes {
    maxGpuCount
    maxGpuCount
    maxGpuCountCommunityCloud
    maxGpuCountSecureCloud
    minPodGpuCount
    id
    displayName
    memoryInGb
    secureCloud
    communityCloud
    __typename
  }
}
"""

query_pod_types = """
query GpuTypes($lowestPriceInput: GpuLowestPriceInput, $gpuTypesInput: GpuTypeFilter) {
  gpuTypes(input: $gpuTypesInput) {
    lowestPrice(input: $lowestPriceInput) {
      minimumBidPrice
      uninterruptablePrice
      minVcpu
      minMemory
      stockStatus
      compliance
      countryCode
      __typename
    }
    maxGpuCount
    id
    nodeGroupDatacenters {
        id
    }
    clusterPrice
    displayName
    memoryInGb
    securePrice
    secureSpotPrice
    communityPrice
    communitySpotPrice
    oneMonthPrice
    threeMonthPrice
    sixMonthPrice
    secureSpotPrice
    __typename
  }
}
"""



import warnings

import requests

from gpuhunt.providers.vastai import get_dstack_gpu_name, get_vastai_gpu_names


def test_real_world_vastai_offers():
    """Test that GPU name conversion works correctly for all real-world offers from Vast.ai."""
    # Get all offers from Vast.ai
    response = requests.post(
        "https://cloud.vast.ai/api/v0/bundles/", json={"limit": 3000}, timeout=10
    )
    response.raise_for_status()
    offers = response.json()["offers"]

    # Track unique GPU names and their conversions
    unique_gpu_names = set()
    conversion_issues = set()

    for offer in offers:
        vastai_gpu_name = offer["gpu_name"]
        if not vastai_gpu_name:
            continue

        unique_gpu_names.add(vastai_gpu_name)

        # Convert to dstack format and back
        dstack_name = get_dstack_gpu_name(vastai_gpu_name)
        vastai_names = get_vastai_gpu_names(dstack_name)

        # Check if the original name is in the converted back list
        if vastai_gpu_name not in vastai_names:
            conversion_issues.add(vastai_gpu_name)

    # Print statistics about mismatched GPUs
    if conversion_issues:
        warning_msg = f"Found {len(conversion_issues)} GPU names without valid mapping:\n"
        for gpu_name in sorted(conversion_issues):
            warning_msg += f"- {gpu_name}\n"
        warnings.warn(warning_msg)



import copy
import enum
import importlib.resources
import json
import logging
import re
from collections import defaultdict, namedtuple
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from typing import Optional, cast

import google.cloud.billing_v1 as billing_v1
import google.cloud.compute_v1 as compute_v1
from google.cloud import tpu_v2
from google.cloud.billing_v1 import CloudCatalogClient, ListSkusRequest
from google.cloud.billing_v1.types.cloud_catalog import Sku
from google.cloud.location import locations_pb2
from typing_extensions import NotRequired, TypedDict

from gpuhunt._internal.models import AcceleratorVendor, QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
compute_service = "services/6F81-5844-456A"
AcceleratorDetails = namedtuple("AcceleratorDetails", ["name", "memory"])
# As of 2024-14-08, this mapping contains only Nvidia accelerators; update gpu_vendor
# inferring code in fill_gpu_vendors_and_names() if a non-Nvidia accelerator is added
accelerator_details = {
    "nvidia-b200": AcceleratorDetails("B200", 180.0),
    "nvidia-a100-80gb": AcceleratorDetails("A100", 80.0),
    "nvidia-h100-80gb": AcceleratorDetails("H100", 80.0),
    "nvidia-h100-mega-80gb": AcceleratorDetails("H100", 80.0),
    "nvidia-l4": AcceleratorDetails("L4", 24.0),
    "nvidia-tesla-a100": AcceleratorDetails("A100", 40.0),
    "nvidia-tesla-k80": AcceleratorDetails("K80", 12.0),
    "nvidia-tesla-p100": AcceleratorDetails("P100", 16.0),
    "nvidia-tesla-p4": AcceleratorDetails("P4", 8.0),
    "nvidia-tesla-t4": AcceleratorDetails("T4", 16.0),
    "nvidia-tesla-v100": AcceleratorDetails("V100", 16.0),
    "nvidia-rtx-pro-6000": AcceleratorDetails("RTXPRO6000", 96.0),
}
CpuMemory = namedtuple("CpuMemory", ["cpu", "memory"])
accelerator_limits = {
    "nvidia-tesla-k80": [
        CpuMemory(8, 52),
        CpuMemory(16, 104),
        CpuMemory(32, 208),
        CpuMemory(64, 208),
    ],
    "nvidia-tesla-p100": [CpuMemory(16, 104), CpuMemory(32, 208), CpuMemory(96, 624)],
    "nvidia-tesla-p4": [CpuMemory(24, 156), CpuMemory(48, 312), CpuMemory(96, 624)],
    "nvidia-tesla-t4": [CpuMemory(48, 312), CpuMemory(48, 312), CpuMemory(96, 624)],
    "nvidia-tesla-v100": [
        CpuMemory(12, 78),
        CpuMemory(24, 156),
        CpuMemory(48, 312),
        CpuMemory(96, 624),
    ],
}
accelerator_counts = [1, 2, 4, 8, 16]
hours_in_month = 730  # according to GCP pricing
multi_token_vm_families = ["a3-megagpu"]

# https://cloud.google.com/compute/docs/disks/local-ssd#lssd_disks_fixed
local_ssd_sizes_gib = {
    "c3-standard-4-lssd": 1 * 375,
    "c3-standard-8-lssd": 2 * 375,
    "c3-standard-22-lssd": 4 * 375,
    "c3-standard-44-lssd": 8 * 375,
    "c3-standard-88-lssd": 16 * 375,
    "c3-standard-176-lssd": 32 * 375,
    "c3d-standard-8-lssd": 1 * 375,
    "c3d-standard-16-lssd": 1 * 375,
    "c3d-standard-30-lssd": 2 * 375,
    "c3d-standard-60-lssd": 4 * 375,
    "c3d-standard-90-lssd": 8 * 375,
    "c3d-standard-180-lssd": 16 * 375,
    "c3d-standard-360-lssd": 32 * 375,
    "c3d-highmem-8-lssd": 1 * 375,
    "c3d-highmem-16-lssd": 1 * 375,
    "c3d-highmem-30-lssd": 2 * 375,
    "c3d-highmem-60-lssd": 4 * 375,
    "c3d-highmem-90-lssd": 8 * 375,
    "c3d-highmem-180-lssd": 16 * 375,
    "c3d-highmem-360-lssd": 32 * 375,
    "a3-megagpu-8g": 16 * 375,
    "a3-highgpu-8g": 16 * 375,
    "a2-ultragpu-1g": 1 * 375,
    "a2-ultragpu-2g": 2 * 375,
    "a2-ultragpu-4g": 4 * 375,
    "a2-ultragpu-8g": 8 * 375,
    "z3-standard-88-lssd": 12 * 3000,
    "z3-standard-176-lssd": 12 * 3000,
}


@dataclass
class TPUHardwareSpec:
    name: str
    cpu: int
    memory_gb: int
    hbm_gb: int


TPU_HARDWARE_SPECS = [
    TPUHardwareSpec(name="v2-8", cpu=96, memory_gb=334, hbm_gb=64),
    TPUHardwareSpec(name="v3-8", cpu=96, memory_gb=334, hbm_gb=128),
    TPUHardwareSpec(name="v5litepod-1", cpu=24, memory_gb=48, hbm_gb=16),
    TPUHardwareSpec(name="v5litepod-2", cpu=112, memory_gb=192, hbm_gb=16),
    TPUHardwareSpec(name="v5litepod-8", cpu=224, memory_gb=384, hbm_gb=128),
    TPUHardwareSpec(name="v5p-8", cpu=208, memory_gb=448, hbm_gb=95),
    TPUHardwareSpec(name="v6e-1", cpu=44, memory_gb=176, hbm_gb=32),
    TPUHardwareSpec(name="v6e-4", cpu=180, memory_gb=720, hbm_gb=128),
    TPUHardwareSpec(name="v6e-8", cpu=180, memory_gb=1440, hbm_gb=256),
]


# For newer TPUs, the specs are described in the docs: https://cloud.google.com/tpu/docs/v6e
# For older TPUs, the specs are collected manually from running instances.
TPU_HARDWARE_SPECS = [
    TPUHardwareSpec(name="v2-8", cpu=96, memory_gb=334, hbm_gb=64),
    TPUHardwareSpec(name="v3-8", cpu=96, memory_gb=334, hbm_gb=128),
    TPUHardwareSpec(name="v5litepod-1", cpu=24, memory_gb=48, hbm_gb=16),
    TPUHardwareSpec(name="v5litepod-4", cpu=112, memory_gb=192, hbm_gb=64),
    TPUHardwareSpec(name="v5litepod-8", cpu=224, memory_gb=384, hbm_gb=128),
    TPUHardwareSpec(name="v5p-8", cpu=208, memory_gb=448, hbm_gb=95),
    TPUHardwareSpec(name="v6e-1", cpu=44, memory_gb=176, hbm_gb=32),
    TPUHardwareSpec(name="v6e-4", cpu=180, memory_gb=720, hbm_gb=128),
    TPUHardwareSpec(name="v6e-8", cpu=180, memory_gb=1440, hbm_gb=256),
]
TPU_NAME_TO_HARDWARE_SPEC = {spec.name: spec for spec in TPU_HARDWARE_SPECS}


def load_tpu_pricing() -> dict:
    return json.loads(
        importlib.resources.files("gpuhunt.resources").joinpath("tpu_pricing.json").read_text()
    )


# A manually filled TPU pricing table from the pricing page.
# On-demand TPUs - https://cloud.google.com/tpu/pricing?hl=en.
# Spot TPUs - https://cloud.google.com/spot-vms/pricing?hl=en.
# It's needed since the TPU pricing API does not return prices for all regions.
# The API may also return 1-year Commitment prices instead of on-demand prices.
TPU_PRICING_TABLE = load_tpu_pricing()


class GCPProvider(AbstractProvider):
    NAME = "gcp"

    def __init__(self, project: str):
        # todo credentials
        self.project = project
        self.machine_types_client = compute_v1.MachineTypesClient()
        self.accelerator_types_client = compute_v1.AcceleratorTypesClient()
        self.regions_client = compute_v1.RegionsClient()
        self.cloud_catalog_client = billing_v1.CloudCatalogClient()

    def list_preconfigured_instances(self) -> list[RawCatalogItem]:
        def _list_zone_instances(zone: str) -> list[RawCatalogItem]:
            zone_instances = []
            logger.info("Fetching instances for zone %s", zone)
            for machine_type in self.machine_types_client.list(project=self.project, zone=zone):
                if machine_type.deprecated.state == compute_v1.DeprecationStatus.State.DEPRECATED:
                    continue
                gpu = None
                if machine_type.accelerators:
                    accelerator = machine_type.accelerators[0].guest_accelerator_type
                    gpu = accelerator_details.get(accelerator)
                    if gpu is None:
                        logger.warning("Unknown accelerator type: %s", accelerator)
                        continue

                instance = RawCatalogItem(
                    instance_name=machine_type.name,
                    location=zone,
                    cpu=machine_type.guest_cpus,
                    memory=round(machine_type.memory_mb / 1024, 1),
                    gpu_count=(machine_type.accelerators[0].guest_accelerator_count if gpu else 0),
                    # gpu_name is canonicalized and gpu_vendor is set later
                    # in fill_gpu_vendors_and_names(), for now we use AcceleratorType.name
                    # as a name (it contains a vendor prefix like "nvidia-")
                    gpu_name=(
                        machine_type.accelerators[0].guest_accelerator_type if gpu else None
                    ),
                    gpu_vendor=None,
                    gpu_memory=gpu.memory if gpu else None,
                    price=None,
                    spot=None,
                    disk_size=None,
                )
                zone_instances.append(instance)
            return zone_instances

        instances = []
        futures = []
        with ThreadPoolExecutor(max_workers=8) as ex:
            for region in self.regions_client.list(project=self.project):
                for zone_url in region.zones:
                    zone = zone_url.split("/")[-1]
                    futures.append(ex.submit(_list_zone_instances, zone))
        for future in as_completed(futures):
            instances.extend(future.result())
        return instances

    def add_gpus(self, instances: list[RawCatalogItem]):
        def _list_zone_instances(
            zone: str, zone_n1_instances: list[RawCatalogItem]
        ) -> list[RawCatalogItem]:
            logger.info("Fetching GPUs for zone %s", zone)
            zone_instances = []
            for accelerator in self.accelerator_types_client.list(project=self.project, zone=zone):
                if accelerator.name not in accelerator_limits:
                    continue
                for n, limit in zip(accelerator_counts, accelerator_limits[accelerator.name]):
                    for instance in zone_n1_instances:
                        if instance.cpu > limit.cpu or instance.memory > limit.memory:
                            continue
                        i = copy.deepcopy(instance)
                        i.gpu_count = n
                        i.gpu_name = accelerator.name
                        i.gpu_memory = accelerator_details[accelerator.name].memory
                        zone_instances.append(i)
            return zone_instances

        n1_instances = defaultdict(list)
        for instance in instances:
            if instance.instance_name.startswith("n1-"):
                n1_instances[instance.location].append(instance)

        instances_with_gpus = []
        futures = []
        with ThreadPoolExecutor(max_workers=8) as ex:
            for zone, zone_n1_instances in n1_instances.items():
                futures.append(ex.submit(_list_zone_instances, zone, zone_n1_instances))
        for future in as_completed(futures):
            instances_with_gpus.extend(future.result())
        instances += instances_with_gpus

    def fill_prices(self, instances: list[RawCatalogItem]) -> list[RawCatalogItem]:
        logger.info("Fetching prices")
        skus = self.cloud_catalog_client.list_skus(parent=compute_service)
        prices = Prices()
        prices.add_skus(skus)

        offers = []
        for instance in instances:
            for capacity_type in CapacityType:
                price = prices.get_instance_price(instance, capacity_type)
                if price is None:
                    continue

                offer = copy.deepcopy(instance)
                offer.price = round(price, 6)
                offer.spot = capacity_type is CapacityType.SPOT
                cast(GCPCatalogItemProviderData, offer.provider_data)["is_dws_calendar_mode"] = (
                    capacity_type is CapacityType.DWS_CALENDAR_MODE
                )
                offers.append(offer)
        return offers

    def fill_gpu_vendors_and_names(self, offers: list[RawCatalogItem]) -> None:
        # Modifies offers in the list in-place
        for offer in offers:
            accelerator_type = offer.gpu_name
            if not accelerator_type:
                continue
            offer.gpu_name = accelerator_details[accelerator_type].name
            if accelerator_type.startswith("nvidia-"):
                offer.gpu_vendor = AcceleratorVendor.NVIDIA.value
            else:
                logger.warning("Unknown accelerator vendor: %s", accelerator_type)

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        instances = self.list_preconfigured_instances()
        self.add_gpus(instances)
        offers = self.fill_prices(instances)
        self.fill_gpu_vendors_and_names(offers)
        offers.extend(get_tpu_offers(self.project))
        set_flags(offers)
        offers = add_legacy_g4_preview(offers)
        return sorted(offers, key=lambda i: i.price)

    @classmethod
    def filter(cls, offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
        return [
            i
            for i in offers
            if (
                any(
                    i.instance_name.startswith(family)
                    for family in [
                        "m4-",
                        "c4-",
                        "n4-",
                        "h3-",
                        "n2-",
                        "e2-medium",
                        "e2-standard-",
                        "e2-highmem-",
                        "e2-highcpu-",
                        "m1-",
                        "a2-",
                        "g2-",
                    ]
                )
                or (i.gpu_name and i.gpu_name not in ["K80", "P4"])
            )
            and not (
                # Filter out on-demand offers that are not actually available on demand.
                # https://cloud.google.com/compute/docs/accelerator-optimized-machines#consumption_option_availability_by_machine_type
                i.spot == False
                and not cast(GCPCatalogItemProviderData, i.provider_data).get(
                    "is_dws_calendar_mode"
                )
                and (
                    i.instance_name.startswith("a4x-")
                    or i.instance_name.startswith("a4-")
                    or i.instance_name.startswith("a3-ultragpu-")
                    or (
                        i.instance_name.startswith("a3-highgpu-")
                        and (i.gpu_count is None or i.gpu_count < 8)
                    )
                )
            )
        ]


class GCPCatalogItemProviderData(TypedDict):
    is_dws_calendar_mode: NotRequired[bool]


class CapacityType(enum.Enum):
    ON_DEMAND = enum.auto()
    SPOT = enum.auto()
    DWS_CALENDAR_MODE = enum.auto()


RegionCapacityType = tuple[str, CapacityType]
PricePerRegionCapacityType = dict[RegionCapacityType, float]


class Prices:
    def __init__(self):
        self.cpu: defaultdict[str, PricePerRegionCapacityType] = defaultdict(dict)
        self.gpu: defaultdict[str, PricePerRegionCapacityType] = defaultdict(dict)
        self.ram: defaultdict[str, PricePerRegionCapacityType] = defaultdict(dict)
        self.local_ssd: PricePerRegionCapacityType = dict()
        self.gpu_slice: defaultdict[str, PricePerRegionCapacityType] = defaultdict(dict)

    def add_skus(self, skus: Iterable[Sku]) -> None:
        for sku in skus:
            if sku.category.usage_type not in ["OnDemand", "Preemptible"]:
                continue
            if any(
                word in sku.description
                for word in [
                    "Sole Tenancy",
                    "Reserved",
                    "Premium",
                    "Custom",
                    "suspended",
                ]
            ):
                continue
            if sku.category.resource_family == "Compute":
                self.add_compute_sku(sku)
            elif sku.category.resource_family == "Storage":
                self.add_storage_sku(sku)

    def add_compute_sku(self, sku: Sku) -> None:
        if "(1 gpu slice)" in sku.description:
            self.add_compute_gpu_slice_sku(sku)
            return

        if "RTX 6000 96GB" in sku.description:
            family = "nvidia-rtx-pro-6000"
            resource = "gpu"
        else:
            r = re.match(
                r"^(?:spot preemptible )?(.+) (gpu|ram|core)",
                sku.description,
                flags=re.IGNORECASE,
            )
            if not r:
                return
            family, resource = r.groups()
            resource = resource.lower()

        if resource == "gpu":
            family = family.replace(" ", "-").lower()
            family = {
                "nvidia-tesla-a100-80gb": "nvidia-a100-80gb",
                "nvidia-h100-80gb-mega": "nvidia-h100-mega-80gb",
                "nvidia-h100-80gb-plus": "nvidia-h100-mega-80gb",
            }.get(family, family)
        else:
            r = re.match(r"^([a-z]\d.?) ", family.lower())
            if r:
                family = r.group(1)
            else:
                family = {
                    "Memory-optimized Instance": "m1",
                    "Compute optimized Instance": "c2",
                    "Compute optimized": "c2",
                    "A3Plus Instance": "a3-megagpu",
                }.get(family, family)

        price = self._calculate_sku_price(sku)
        resource_prices = {
            "core": self.cpu,
            "gpu": self.gpu,
            "ram": self.ram,
        }[resource]
        self._add_price(sku, resource_prices[family], price)

    def add_compute_gpu_slice_sku(self, sku: Sku) -> None:
        if sku.description.startswith(
            "Spot Preemptible A4 Nvidia B200"
        ) or sku.description.startswith("DWS Calendar Mode A4 Nvidia B200"):
            gpu = "nvidia-b200"
        else:
            return
        price = self._calculate_sku_price(sku)
        self._add_price(sku, self.gpu_slice[gpu], price)

    def add_storage_sku(self, sku: Sku) -> None:
        if sku.description.lower().startswith("ssd backed local storage"):
            price = self._calculate_sku_price(sku) / hours_in_month
            self._add_price(sku, self.local_ssd, price)

    @staticmethod
    def _calculate_sku_price(sku: Sku) -> float:
        price = sku.pricing_info[0].pricing_expression.tiered_rates[0].unit_price
        return price.units + price.nanos / 1e9

    @staticmethod
    def _add_price(sku: Sku, family_prices: PricePerRegionCapacityType, price: float) -> None:
        if sku.category.usage_type == "Preemptible":
            capacity_type = CapacityType.SPOT
        elif "DWS Calendar Mode" in sku.description:
            capacity_type = CapacityType.DWS_CALENDAR_MODE
        else:
            capacity_type = CapacityType.ON_DEMAND
        for region in sku.service_regions:
            family_prices[(region, capacity_type)] = price

    def get_instance_price(
        self, instance: RawCatalogItem, capacity_type: CapacityType
    ) -> Optional[float]:
        vm_family = self.get_vm_family(instance.instance_name)
        if vm_family in ["g1", "f1", "m2"]:  # shared-core and reservation-only
            return None

        region_capacity_type = (instance.location[:-2], capacity_type)

        # For some instances, the price is proportional to the number of GPUs
        if instance.gpu_name and region_capacity_type in self.gpu_slice[instance.gpu_name]:
            return instance.gpu_count * self.gpu_slice[instance.gpu_name][region_capacity_type]

        # For others, the price consists of several components
        price = 0
        if (
            region_capacity_type not in self.cpu[vm_family]
            or region_capacity_type not in self.ram[vm_family]
        ):
            return None
        price += instance.cpu * self.cpu[vm_family][region_capacity_type]
        price += instance.memory * self.ram[vm_family][region_capacity_type]
        if instance.gpu_name:
            if region_capacity_type not in self.gpu[instance.gpu_name]:
                return None
            price += instance.gpu_count * self.gpu[instance.gpu_name][region_capacity_type]
        if instance.instance_name in local_ssd_sizes_gib:
            price += (
                local_ssd_sizes_gib[instance.instance_name] * self.local_ssd[region_capacity_type]
            )

        return price

    @staticmethod
    def get_vm_family(instance_name: str) -> str:
        for family in multi_token_vm_families:
            if instance_name.startswith(family):
                return family
        return instance_name.split("-")[0]


def set_flags(catalog_items: list[RawCatalogItem]) -> None:
    for item in catalog_items:
        if cast(GCPCatalogItemProviderData, item.provider_data).get("is_dws_calendar_mode"):
            item.flags.append("gcp-dws-calendar-mode")
        if item.instance_name.startswith("a4-"):
            item.flags.append("gcp-a4")
        elif item.instance_name.startswith("g4-standard-"):
            item.flags.append("gcp-g4")


# TODO: drop when dstack 0.19.33 is no longer relevant
def add_legacy_g4_preview(catalog_items: list[RawCatalogItem]) -> list[RawCatalogItem]:
    """
    For each g4-standard-* instance, add a duplicate item with the "gcp-g4-preview" flag.

    This is only needed for dstack 0.19.33, where the flag "gcp-g4-preview"
    is used instead of "gcp-g4".
    """
    new_items = []
    for item in catalog_items:
        new_items.append(item)
        if item.instance_name.startswith("g4-standard-"):
            preview_item = copy.deepcopy(item)
            preview_item.flags.remove("gcp-g4")
            preview_item.flags.append("gcp-g4-preview")
            new_items.append(preview_item)
    return new_items


def get_tpu_offers(project_id: str) -> list[RawCatalogItem]:
    logger.info("Fetching TPU offers")
    raw_catalog_items: list[RawCatalogItem] = []
    catalog_items: list[dict] = get_catalog_items(project_id)
    # For some TPU offers in some regions, GCP does not list prices at all. Skip such offers.
    filtered_catalog_items = [item for item in catalog_items if item["price"] is not None]
    for item in filtered_catalog_items:
        hardware_spec = get_tpu_hardware_spec(item["instance_name"])
        if hardware_spec is None:
            logger.debug("No TPU hardware spec for %s", item["instance_name"])
            continue
        on_demand_item = RawCatalogItem(
            instance_name=item["instance_name"],
            location=item["location"],
            price=item["price"],
            cpu=hardware_spec.cpu,
            memory=hardware_spec.memory_gb,
            gpu_vendor=AcceleratorVendor.GOOGLE.value,
            gpu_count=1,
            gpu_name=item["instance_name"],
            gpu_memory=hardware_spec.hbm_gb,
            spot=False,
            disk_size=None,
        )
        raw_catalog_items.append(on_demand_item)
        if item["spot"]:
            spot_item = copy.deepcopy(on_demand_item)
            spot_item.price = item["spot"]
            spot_item.spot = True
            raw_catalog_items.append(spot_item)
    return raw_catalog_items


def get_catalog_items(project_id: str) -> list[dict]:
    """
    Returns TPU configurations with pricing info.
    Each configuration contains on-demand price and spot price but any price can be missing.
    This is because the API does not return prices for all regions.
    As a backup, the prices are taken from the pricing table on the GCP website,
    but it also does not contain all the prices.
    Even when creating some TPUs in some regions via the GCP console,
    the price is not shown (e.g. v6e in us-south1).
    """
    tpu_prices: list[dict] = get_tpu_prices()
    configs: list[dict] = get_tpu_configs(project_id)
    for config in configs:
        instance_name = config["instance_name"]
        location = config["location"].rsplit("-", 1)[0]  # Remove the part after the last '-'
        no_of_chips = config["no_of_chips"]
        tpu_version, no_of_cores = instance_name.rsplit("-", 1)
        no_of_cores = int(no_of_cores)
        if tpu_version in ["v5litepod", "v5p", "v6e"]:
            # For TPU-v5 series, the API provides per chip price.
            on_demand_base_price = find_base_price_v5(
                tpu_version, location, tpu_prices, spot=False
            )
            if on_demand_base_price is not None:
                on_demand_price = on_demand_base_price * no_of_chips
            else:
                on_demand_price = find_tpu_price_static_src(
                    tpu_version, no_of_cores, location, no_of_chips, False
                )
            spot_base_price = find_base_price_v5(tpu_version, location, tpu_prices, spot=True)
            if spot_base_price is not None:
                spot_price = spot_base_price * no_of_chips
            else:
                spot_price = find_tpu_price_static_src(
                    tpu_version, no_of_cores, location, no_of_chips, True
                )
        elif tpu_version in ["v2", "v3", "v4"]:
            # For TPU-v2 and TPU-v3, the pricing API provides the prices of 8 TPU cores.
            # For TPU-v4, the API provides the price of TPU-v4 pods.
            if no_of_cores > 8 or tpu_version == "v4":
                base_instance_name = f"{tpu_version}-8"
                base_no_of_chips = find_no_of_chips(base_instance_name, configs)
                on_demand_base_price = find_base_price(
                    tpu_version, location, tpu_prices, spot=False, is_pod=True
                )
                if on_demand_base_price is not None and base_no_of_chips is not None:
                    on_demand_price = (on_demand_base_price / base_no_of_chips) * no_of_chips
                else:
                    on_demand_price = find_tpu_price_static_src(
                        tpu_version, no_of_cores, location, no_of_chips, False
                    )
                spot_base_price = find_base_price(
                    tpu_version, location, tpu_prices, spot=True, is_pod=True
                )
                if spot_base_price is not None and base_no_of_chips is not None:
                    spot_price = (spot_base_price / base_no_of_chips) * no_of_chips
                else:
                    spot_price = find_tpu_price_static_src(
                        tpu_version, no_of_cores, location, no_of_chips, True
                    )
            elif no_of_cores == 8:
                on_demand_price = find_base_price(
                    tpu_version, location, tpu_prices, spot=False, is_pod=False
                )
                spot_price = find_base_price(
                    tpu_version, location, tpu_prices, spot=True, is_pod=False
                )
            else:
                logger.warning("Unknown TPU type %s", instance_name)
                continue
        else:
            logger.warning("Unknown TPU version %s. Skipping offer.", tpu_version)
            continue
        if on_demand_price is None:
            logger.debug("Failed to find on-demand price for %s in %s", instance_name, location)
        if spot_price is None:
            logger.debug("Failed to find spot price for %s in %s", instance_name, location)
        config["price"] = on_demand_price
        config["spot"] = spot_price
    return configs


def get_tpu_prices() -> list[dict]:
    client = CloudCatalogClient()
    tpu_configs = []
    # E000-3F24-B8AA contains prices for TPU versions v2,v3,v4.
    # 6F81-5844-456A contains prices for newer TPU versions v5p, v5litepod(v5e), v6e.
    service_names = ["services/E000-3F24-B8AA", "services/6F81-5844-456A"]
    for service_name in service_names:
        request = ListSkusRequest(parent=service_name)
        response = client.list_skus(request=request)
        for sku in response.skus:
            if sku.category.resource_group != "TPU":
                continue
            if sku.category.usage_type not in ["OnDemand", "Preemptible"]:
                continue
            tpu_version = extract_tpu_version(sku.description)
            if tpu_version:
                is_pod = True if "Pod" in sku.description else False
                spot = True if "Preemptible" in sku.description else False
                price = sku.pricing_info[0].pricing_expression.tiered_rates[0].unit_price
                price = price.units + price.nanos / 1e9
                tpu_configs.append(
                    {
                        "instance_name": tpu_version,
                        "is_pod": is_pod,
                        "spot": spot,
                        "regions": sku.service_regions,
                        "price": price,
                        "description": sku.description,
                    }
                )
    return tpu_configs


def find_base_price(
    tpu_version: str, location: str, tpu_prices: list[dict], spot: bool, is_pod: bool
) -> Optional[float]:
    for price_info in tpu_prices:
        if (
            price_info["instance_name"] == tpu_version
            and any(loc.startswith(location) for loc in price_info["regions"])
            and price_info["spot"] == spot
            and price_info["is_pod"] == is_pod
        ):
            return price_info["price"]
    return None


def find_no_of_chips(instance_name: str, configs: list[dict]):
    for config in configs:
        if config["instance_name"] == instance_name:
            return config["no_of_chips"]
    return None


def find_tpu_price_static_src(
    tpu_version: str, num_cores: int, tpu_region: str, no_of_chips: int, spot: bool
) -> Optional[float]:
    # The pricing page names v5litepod as v5e
    tpu_version = "v5e" if tpu_version == "v5litepod" else tpu_version
    # The pricing page lists different (device and pod) prices per chip for v2 and v3.
    # The device is the smallest configuration with 8 TPU cores (e.g. v3-8).
    # TPU Pod connects mulitple TPU devices (e.g. v3-32).
    # Not applicable for newer generations.
    tpu_type = f"TPU {tpu_version}"
    if tpu_version in ["v2", "v3", "v4"]:
        is_pod = num_cores > 8 or tpu_version == "v4"
        tpu_type = f"TPU {tpu_version} pod" if is_pod else f"TPU {tpu_version} device"
    price_key = "On Demand (USD)"
    if spot:
        price_key = "Spot (USD)"
    try:
        return TPU_PRICING_TABLE[tpu_type][tpu_region][price_key] * no_of_chips
    except KeyError:
        logger.debug(f"KeyError for {tpu_type} {tpu_region} {price_key}")
        return None


def find_base_price_v5(
    tpu_version: str, location: str, tpu_prices: list[dict], spot: bool
) -> Optional[float]:
    for price_info in tpu_prices:
        if (
            price_info["instance_name"] == tpu_version
            and any(loc.startswith(location) for loc in price_info["regions"])
            and price_info["spot"] == spot
        ):
            return price_info["price"]
    return None


def get_tpu_configs(project_id: str) -> list[dict]:
    def _list_zone_configs(zone: str) -> list[dict]:
        zone_instances = []
        if zone in ["us-east1-b"]:
            # These zones return
            # google.api_core.exceptions.ServiceUnavailable: 503 502:Bad Gateway
            return []
        parent = f"projects/{project_id}/locations/{zone}"
        request = tpu_v2.ListAcceleratorTypesRequest(
            parent=parent,
        )
        page_result = client.list_accelerator_types(request=request)
        for response in page_result:
            no_of_chips = get_no_of_chips(response.accelerator_configs[0].topology)
            zone_instances.append(
                {
                    "instance_name": response.type_,
                    "location": zone,
                    "no_of_chips": no_of_chips,
                }
            )
        return zone_instances

    client = tpu_v2.TpuClient()
    instances: list[dict] = []
    futures = []
    with ThreadPoolExecutor(max_workers=8) as ex:
        for zone in get_locations(project_id):
            futures.append(ex.submit(_list_zone_configs, zone))
    for future in as_completed(futures):
        instances.extend(future.result())
    return instances


def get_no_of_chips(expression: str) -> int:
    factors = expression.split("x")
    factors = map(int, factors)
    product = 1
    for factor in factors:
        product *= factor
    return product


def get_locations(project_id: str) -> list[str]:
    client = tpu_v2.TpuClient()
    parent = f"projects/{project_id}"
    list_locations_request = client.list_locations(locations_pb2.ListLocationsRequest(name=parent))
    locations = [loc.location_id for loc in list_locations_request.locations]
    # TPU V4 only available in location us-central2-b only.
    # us-central2-b needs to be enabled in the project.
    return locations


def extract_tpu_version(input_string: str) -> Optional[str]:
    # The regular expression pattern to find a substring starting with 'Tpu'
    pattern = r"\bTpu[-\w]*\b"
    # Search for the first match of the pattern
    match = re.search(pattern, input_string, re.IGNORECASE)
    if match:
        tpu_match = match.group().lower()
        # The regular expression pattern to find the version part
        version_pattern = r"v\d+[a-z]*"
        version_match = re.search(version_pattern, tpu_match)
        if version_match:
            # Name of v5e in gcp console is v5litepod
            version = "v5litepod" if version_match.group() == "v5e" else version_match.group()
            return version
    return None


def get_tpu_hardware_spec(instance_name: str) -> Optional[TPUHardwareSpec]:
    return TPU_NAME_TO_HARDWARE_SPEC.get(instance_name)



import logging
import math
from collections import namedtuple
from itertools import chain
from math import ceil
from typing import Optional, TypeVar, Union

import requests

from gpuhunt import QueryFilter, RawCatalogItem
from gpuhunt._internal.constraints import (
    find_accelerators,
    get_compute_capability,
    is_between,
)
from gpuhunt._internal.models import AcceleratorVendor
from gpuhunt.providers import AbstractProvider

CpuMemoryGpu = namedtuple("CpuMemoryGpu", ["cpu", "memory", "gpu"])
logger = logging.getLogger(__name__)

API_URL = "https://rest.compute.cudo.org/v1"
MIN_CPU = 2
MIN_MEMORY = 8
RAM_PER_VRAM = 2
RAM_DIV = 2
CPU_DIV = 2
RAM_PER_CORE = 4
MIN_DISK_SIZE = 100


class CudoProvider(AbstractProvider):
    NAME = "cudo"

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        offers = self.fetch_offers(query_filter, balance_resources)
        offers = get_min_price_for_location_and_instance(offers)
        return sorted(offers, key=lambda i: i.price)

    def fetch_offers(
        self, query_filter: Optional[QueryFilter], balance_resources
    ) -> list[RawCatalogItem]:
        machine_types = self.list_vm_machine_types()
        if query_filter is not None:
            return self.optimize_offers(machine_types, query_filter, balance_resources)
        else:
            offers = []
            for machine_type in machine_types:
                gpu_model = machine_type["gpuModel"]
                gpu_model_name = gpu_name(gpu_model)
                if gpu_model_name is None:
                    if gpu_model:
                        logger.warning("Failed to get GPU name from gpuModel: '%s'", gpu_model)
                    continue
                gpu_memory_size = get_memory(gpu_model_name)
                if gpu_memory_size is None:
                    continue
                machine_type["gpu_name"] = gpu_model_name
                machine_type["gpu_memory"] = gpu_memory_size
                optimized_specs = optimize_offers_with_gpu(
                    QueryFilter(), machine_type, balance_resources=False
                )
                raw_catalogs = [get_raw_catalog(machine_type, spec) for spec in optimized_specs]
                offers.append(raw_catalogs)
            return list(chain.from_iterable(offers))

    @staticmethod
    def list_vm_machine_types() -> dict:
        resp = requests.request(
            method="GET",
            url=f"{API_URL}/vms/machine-types",
            timeout=10,
        )
        if resp.ok:
            data = resp.json()
            return data["machineTypes"]
        resp.raise_for_status()

    @staticmethod
    def optimize_offers(
        machine_types, q: QueryFilter, balance_resource: bool
    ) -> list[RawCatalogItem]:
        offers = []

        # Empty query, example: QureyFilter(provider="cudo")
        get_all_offers = all(
            condition is None
            for condition in (
                q.min_gpu_count,
                q.max_gpu_count,
                q.min_total_gpu_memory,
                q.max_total_gpu_memory,
                q.min_gpu_memory,
                q.max_gpu_memory,
                q.gpu_name,
            )
        )

        has_significant_gpu_filter = any(
            condition is not None
            for condition in (
                q.min_gpu_count or None,
                q.max_gpu_count or None,
                q.min_total_gpu_memory or None,
                q.max_total_gpu_memory or None,
                q.min_gpu_memory or None,
                q.max_gpu_memory or None,
                q.gpu_name or None,
            )
        )

        if has_significant_gpu_filter or get_all_offers:
            # filter offers with gpus
            gpu_machine_types = [
                vm for vm in machine_types if vm["maxGpuFree"] != 0 and vm["gpuModelId"]
            ]
            for machine_type in gpu_machine_types:
                gpu_model = machine_type["gpuModel"]
                gpu_model_name = gpu_name(gpu_model)
                if gpu_model_name is None:
                    if gpu_model:
                        logger.warning("Failed to get GPU name from gpuModel: '%s'", gpu_model)
                    continue
                gpu_memory_size = get_memory(gpu_model_name)
                if gpu_memory_size is None:
                    continue
                machine_type["gpu_name"] = gpu_model_name
                machine_type["gpu_memory"] = gpu_memory_size
                if not is_between(
                    machine_type["gpu_memory"], q.min_gpu_memory, q.max_total_gpu_memory
                ):
                    continue
                if q.gpu_name is not None and machine_type["gpu_name"].lower() not in map(
                    str.lower, q.gpu_name
                ):
                    continue
                cc = get_compute_capability(machine_type["gpu_name"])
                if not cc or not is_between(
                    cc, q.min_compute_capability, q.max_compute_capability
                ):
                    continue
                optimized_specs = optimize_offers_with_gpu(q, machine_type, balance_resource)
                raw_catalogs = [get_raw_catalog(machine_type, spec) for spec in optimized_specs]
                offers.append(raw_catalogs)

        include_cpu_offers = any(
            (
                q.min_gpu_count == 0,
                (q.min_total_gpu_memory == 0) if q.min_total_gpu_memory is not None else False,
                q.min_gpu_memory == 0,
            )
        )

        if not has_significant_gpu_filter or get_all_offers or include_cpu_offers:
            cpu_only_machine_types = [
                vm for vm in machine_types if vm["maxVcpuFree"] != 0 and not vm["gpuModelId"]
            ]
            for machine_type in cpu_only_machine_types:
                optimized_specs = optimize_offers_no_gpu(q, machine_type, balance_resource)
                raw_catalogs = [get_raw_catalog(machine_type, spec) for spec in optimized_specs]
                offers.append(raw_catalogs)

        return list(chain.from_iterable(offers))


def get_raw_catalog(machine_type, spec):
    raw = RawCatalogItem(
        instance_name=machine_type["machineType"],
        location=machine_type["dataCenterId"],
        spot=False,
        price=round(
            math.fsum(
                (
                    float(machine_type["vcpuPriceHr"]["value"]) * spec["cpu"],
                    float(machine_type["memoryGibPriceHr"]["value"]) * spec["memory"],
                    float(machine_type["gpuPriceHr"]["value"]) * spec.get("gpu", 0),
                    float(machine_type["minStorageGibPriceHr"]["value"]) * spec["disk_size"],
                    float(machine_type["ipv4PriceHr"]["value"]),
                )
            ),
            5,
        ),
        cpu=spec["cpu"],
        memory=spec["memory"],
        gpu_vendor=None,
        gpu_count=spec.get("gpu", 0),
        gpu_name=machine_type.get("gpu_name", ""),
        gpu_memory=machine_type.get("gpu_memory", 0),
        disk_size=spec["disk_size"],
    )
    return raw


def get_min_price_for_location_and_instance(offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
    """
    Returns offers with the minimum price for each unique combination
    of location and instance name.
    """
    min_price_offers = {}
    for offer in offers:
        key = (offer.location, offer.instance_name)
        if key not in min_price_offers or offer.price < min_price_offers[key].price:
            min_price_offers[key] = offer
    return list(min_price_offers.values())


def optimize_offers_with_gpu(
    q: QueryFilter, machine_type: dict, balance_resources: bool
) -> list[dict]:
    # Generate ranges for CPU, GPU, and memory based on the specified minimums, maximums, and available resources
    cpu_range = get_cpu_range(q.min_cpu, q.max_cpu, machine_type["maxVcpuFree"])
    gpu_range = get_gpu_range(q.min_gpu_count, q.max_gpu_count, machine_type["maxGpuFree"])
    memory_range = get_memory_range(q.min_memory, q.max_memory, machine_type["maxMemoryGibFree"])
    min_vcpu_per_memory_gib = machine_type.get("minVcpuPerMemoryGib", 0)
    max_vcpu_per_memory_gib = machine_type.get("maxVcpuPerMemoryGib", float("inf"))
    min_vcpu_per_gpu = machine_type.get("minVcpuPerGpu", 0)
    max_vcpu_per_gpu = machine_type.get("maxVcpuPerGpu", float("inf"))
    unbalanced_specs = []
    # FIXME: this can be an enormously inefficient nested loop
    for cpu in cpu_range:
        for gpu in gpu_range:
            for memory in memory_range:
                # Check CPU/memory constraints
                if not is_between(
                    cpu, memory * min_vcpu_per_memory_gib, memory * max_vcpu_per_memory_gib
                ):
                    continue

                # Check CPU/GPU constraints
                if gpu > 0:
                    if not is_between(cpu, gpu * min_vcpu_per_gpu, gpu * max_vcpu_per_gpu):
                        continue

                # If all constraints are met, append this combination
                unbalanced_specs.append({"cpu": cpu, "memory": memory, "gpu": gpu})

    # If resource balancing is required, filter combinations to meet the balanced memory requirement
    if balance_resources:
        memory_balanced = [
            spec
            for spec in unbalanced_specs
            if spec["memory"]
            == get_balanced_memory(spec["gpu"], machine_type["gpu_memory"], q.max_memory)
        ]
        if len(memory_balanced) > 0:
            balanced_specs = memory_balanced
        else:
            # If fail to use balanced memory, use max available memory for every gpu num
            gpu_num_to_spec = {}
            for spec in unbalanced_specs:
                gpu_num_to_spec[spec["gpu"]] = spec
            balanced_specs = gpu_num_to_spec.values()
        # Add disk
        balanced_specs = [
            {
                "cpu": spec["cpu"],
                "memory": spec["memory"],
                "gpu": spec["gpu"],
                "disk_size": get_balanced_disk_size(
                    machine_type["maxStorageGibFree"],
                    spec["memory"],
                    spec["gpu"] * machine_type["gpu_memory"],
                    q.max_disk_size,
                    q.min_disk_size,
                ),
            }
            for spec in balanced_specs
        ]
        return balanced_specs

    disk_size = q.min_disk_size if q.min_disk_size is not None else MIN_DISK_SIZE
    # Add disk
    unbalanced_specs = [
        {"cpu": spec["cpu"], "memory": spec["memory"], "gpu": spec["gpu"], "disk_size": disk_size}
        for spec in unbalanced_specs
    ]
    return unbalanced_specs


def optimize_offers_no_gpu(q: QueryFilter, machine_type, balance_resource):
    # Generate ranges for CPU, memory based on the specified minimums, maximums, and available resources
    cpu_range = get_cpu_range(q.min_cpu, q.max_cpu, machine_type["maxVcpuFree"])
    memory_range = get_memory_range(q.min_memory, q.max_memory, machine_type["maxMemoryGibFree"])

    # Cudo Specific Constraints
    min_vcpu_per_memory_gib = machine_type.get("minVcpuPerMemoryGib", 0)
    max_vcpu_per_memory_gib = machine_type.get("maxVcpuPerMemoryGib", float("inf"))

    unbalanced_specs = []
    for cpu in cpu_range:
        for memory in memory_range:
            # Check CPU/memory constraints
            if not is_between(
                cpu, memory * min_vcpu_per_memory_gib, memory * max_vcpu_per_memory_gib
            ):
                continue
            # If all constraints are met, append this combination
            unbalanced_specs.append({"cpu": cpu, "memory": memory})

    # If resource balancing is required, filter combinations to meet the balanced memory requirement
    if balance_resource:
        cpu_balanced = [
            spec
            for spec in unbalanced_specs
            if spec["cpu"] == get_balanced_cpu(spec["memory"], q.max_memory)
        ]

        balanced_specs = cpu_balanced
        # Add disk
        disk_size = q.min_disk_size if q.min_disk_size is not None else MIN_DISK_SIZE
        balanced_specs = [
            {"cpu": spec["cpu"], "memory": spec["memory"], "disk_size": disk_size}
            for spec in balanced_specs
        ]
        # Return balanced combinations if any; otherwise, return all combinations
        return balanced_specs

    disk_size = q.min_disk_size if q.min_disk_size is not None else MIN_DISK_SIZE
    # Add disk
    unbalanced_specs = [
        {
            "cpu": spec["cpu"],
            "memory": spec["memory"],
            "gpu": 0,
            "disk_size": min_none(machine_type["maxStorageGibFree"], disk_size),
        }
        for spec in unbalanced_specs
    ]
    return unbalanced_specs


def get_cpu_range(min_cpu, max_cpu, max_cpu_free):
    cpu_range = range(
        min_cpu if min_cpu is not None else MIN_CPU,
        min(max_cpu if max_cpu is not None else max_cpu_free, max_cpu_free) + 1,
    )
    return cpu_range


def get_gpu_range(min_gpu_count, max_gpu_count, max_gpu_free):
    gpu_range = range(
        min_gpu_count if min_gpu_count is not None else 1,
        min(max_gpu_count if max_gpu_count is not None else max_gpu_free, max_gpu_free) + 1,
    )
    return gpu_range


def get_memory_range(min_memory, max_memory, max_memory_gib_free):
    memory_range = range(
        int(min_memory) if min_memory is not None else MIN_MEMORY,
        min(
            int(max_memory) if max_memory is not None else max_memory_gib_free, max_memory_gib_free
        )
        + 1,
    )
    return memory_range


def get_balanced_memory(gpu_count, gpu_memory, max_memory):
    return min_none(
        round_up(RAM_PER_VRAM * gpu_memory * gpu_count, RAM_DIV), round_down(max_memory, RAM_DIV)
    )


def get_balanced_cpu(memory, max_cpu):
    return min_none(
        round_up(ceil(memory / RAM_PER_CORE), CPU_DIV),
        round_down(max_cpu, CPU_DIV),  # can be None
    )


def get_balanced_disk_size(available_disk, memory, total_gpu_memory, max_disk_size, min_disk_size):
    return max_none(
        min_none(
            available_disk,
            max(memory, total_gpu_memory),
            max_disk_size,
        ),
        min_disk_size,
    )


def gpu_name(name: str) -> Optional[str]:
    if not name:
        return None
    result = GPU_MAP.get(name)
    return result


def get_memory(gpu_name: str) -> Optional[int]:
    if not gpu_name:
        return None
    if accelerators := find_accelerators(names=[gpu_name], vendors=[AcceleratorVendor.NVIDIA]):
        return accelerators[0].memory
    return None


def round_up(value: Optional[Union[int, float]], step: int) -> Optional[int]:
    if value is None:
        return None
    return round_down(value + step - 1, step)


def round_down(value: Optional[Union[int, float]], step: int) -> Optional[int]:
    if value is None:
        return None
    return int(value // step * step)


T = TypeVar("T", bound=Union[int, float])


def min_none(*args: Optional[T]) -> T:
    return min(v for v in args if v is not None)


def max_none(*args: Optional[T]) -> T:
    return max(v for v in args if v is not None)


GPU_MAP = {
    "RTX A5000": "A5000",
    "RTX A6000": "A6000",
    "V100": "V100",
    "RTX 3080": "RTX3080",
    "A40 (compute mode)": "A40",
    "L40S (compute mode)": "L40S",
    "A100 80GB PCIe": "A100",
    "H100 SXM": "H100",
    "H100 NVL": "H100NVL",
}



import logging
import re
from dataclasses import dataclass
from typing import Optional, cast

from nebius.aio.channel import Credentials
from nebius.api.nebius.billing.v1alpha1 import (
    CalculatorServiceClient,
    EstimateRequest,
    ResourceSpec,
)
from nebius.api.nebius.common.v1 import ResourceMetadata
from nebius.api.nebius.compute.v1 import (
    CreateInstanceRequest,
    InstanceSpec,
    ListPlatformsRequest,
    ListPlatformsResponse,
    PlatformServiceClient,
    PreemptibleSpec,
    Preset,
    ResourcesSpec,
)
from nebius.api.nebius.iam.v1 import (
    ListProjectsRequest,
    ListTenantsRequest,
    ProjectServiceClient,
    TenantServiceClient,
)
from nebius.sdk import SDK
from typing_extensions import TypedDict

from gpuhunt._internal.constraints import find_accelerators
from gpuhunt._internal.models import (
    AcceleratorInfo,
    AcceleratorVendor,
    JSONObject,
    QueryFilter,
    RawCatalogItem,
)
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
TIMEOUT = 7


@dataclass(frozen=True)
class InfinibandFabric:
    name: str
    platform: str
    region: str


# https://docs.nebius.com/compute/clusters/gpu#fabrics
INFINIBAND_FABRICS = [
    InfinibandFabric("fabric-2", "gpu-h100-sxm", "eu-north1"),
    InfinibandFabric("fabric-3", "gpu-h100-sxm", "eu-north1"),
    InfinibandFabric("fabric-4", "gpu-h100-sxm", "eu-north1"),
    InfinibandFabric("fabric-5", "gpu-h200-sxm", "eu-west1"),
    InfinibandFabric("fabric-6", "gpu-h100-sxm", "eu-north1"),
    InfinibandFabric("fabric-7", "gpu-h200-sxm", "eu-north1"),
    InfinibandFabric("us-central1-a", "gpu-h200-sxm", "us-central1"),
    InfinibandFabric("us-central1-b", "gpu-b200-sxm", "us-central1"),
]


class NebiusProvider(AbstractProvider):
    NAME = "nebius"

    def __init__(self, credentials: Credentials) -> None:
        self.credentials = credentials

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        items: list[RawCatalogItem] = []
        sdk = SDK(credentials=self.credentials)
        calculator = CalculatorServiceClient(sdk)
        try:
            region_to_project_id = get_sample_projects(sdk)
            for region, project_id in region_to_project_id.items():
                platforms = list_platforms(sdk, project_id).items
                for platform in platforms:
                    logger.info("Processing %s/%s", region, platform.metadata.name)
                    gpu = get_gpu_info(platform.metadata.name)
                    for preset in platform.spec.presets:
                        for spot in [False] + (
                            [True] if platform.status.allowed_for_preemptibles else []
                        ):
                            price = get_price(
                                calculator, project_id, platform.metadata.name, preset.name, spot
                            )
                            item = make_item(
                                platform.metadata.name, preset, gpu, region, spot, price
                            )
                            if item is not None:
                                items.append(item)
        finally:
            sdk.sync_close(timeout=TIMEOUT)
        items.sort(key=lambda i: i.price)
        return items


class NebiusCatalogItemProviderData(TypedDict):
    fabrics: list[str]


def get_sample_projects(sdk: SDK) -> dict[str, str]:
    """
    Returns:
        A mapping from region names to project IDs of random projects in those regions.
    """
    tenants = TenantServiceClient(sdk).list(ListTenantsRequest(), per_retry_timeout=TIMEOUT).wait()
    if len(tenants.items) != 1:
        raise ValueError(f"Expected to find 1 tenant, found {(len(tenants.items))}")
    projects = (
        ProjectServiceClient(sdk)
        .list(
            ListProjectsRequest(parent_id=tenants.items[0].metadata.id), per_retry_timeout=TIMEOUT
        )
        .wait()
    )
    region_to_project_id = {}
    for project in projects.items:
        region_to_project_id[project.status.region] = project.metadata.id
    return region_to_project_id


def get_price(
    calculator: CalculatorServiceClient, project_id: str, platform: str, preset: str, spot: bool
) -> float:
    spec = CreateInstanceRequest(
        metadata=ResourceMetadata(parent_id=project_id),
        spec=InstanceSpec(
            resources=ResourcesSpec(platform=platform, preset=preset),
            preemptible=PreemptibleSpec() if spot else None,
        ),
    )
    estimate = calculator.estimate(
        request=EstimateRequest(resource_spec=ResourceSpec(compute_instance_spec=spec))
    ).wait()
    return float(estimate.hourly_cost.general.total.cost)


def list_platforms(sdk: SDK, project_id: str) -> ListPlatformsResponse:
    req = ListPlatformsRequest(
        page_size=999,
        parent_id=project_id,
    )
    return PlatformServiceClient(sdk).list(req, per_retry_timeout=TIMEOUT).wait()


def get_gpu_info(platform: str) -> Optional[AcceleratorInfo]:
    m = re.match(r"gpu-([^-]+)-", platform)
    if m is None:
        return None
    gpu_name = m.group(1)
    accelerator_info = find_accelerators(names=[gpu_name], vendors=[AcceleratorVendor.NVIDIA])
    if len(accelerator_info) != 1:
        return None
    return accelerator_info[0]


def make_item(
    platform: str,
    preset: Preset,
    gpu: Optional[AcceleratorInfo],
    region: str,
    spot: bool,
    price: float,
) -> Optional[RawCatalogItem]:
    fabrics = []
    if preset.allow_gpu_clustering:
        fabrics = [
            f.name for f in INFINIBAND_FABRICS if f.platform == platform and f.region == region
        ]

    item = RawCatalogItem(
        instance_name=f"{platform} {preset.name}",
        location=region,
        price=price,
        cpu=preset.resources.vcpu_count,
        memory=preset.resources.memory_gibibytes,
        gpu_count=0,
        gpu_name=None,
        gpu_memory=None,
        gpu_vendor=None,
        spot=spot,
        disk_size=None,
        provider_data=cast(JSONObject, NebiusCatalogItemProviderData(fabrics=fabrics)),
    )

    if preset.resources.gpu_count:
        if gpu is None:
            logger.warning(
                "Platform %s preset %s has GPUs, but they could not be identified. Skipping",
                platform,
                preset.name,
            )
            return None
        item.gpu_count = preset.resources.gpu_count
        item.gpu_name = gpu.name
        item.gpu_memory = float(gpu.memory)
        item.gpu_vendor = gpu.vendor

    return item



import logging
import os
from typing import Optional

import requests

from gpuhunt._internal.constraints import get_gpu_vendor
from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)

# DigitalOcean Default API endpoints
STANDARD_CLOUD_API_URL = "https://api.digitalocean.com"


class DigitalOceanProvider(AbstractProvider):
    NAME = "digitalocean"

    def __init__(self, api_key: Optional[str] = None, api_url: Optional[str] = None):
        self.api_key = api_key or os.getenv("DIGITAL_OCEAN_API_KEY")
        if not self.api_key:
            raise ValueError("Set the DIGITAL_OCEAN_API_KEY environment variable.")

        self.api_url = api_url or os.getenv("DIGITAL_OCEAN_API_URL", STANDARD_CLOUD_API_URL)

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        offers = self.fetch_offers()
        return sorted(offers, key=lambda i: i.price)

    def fetch_offers(self) -> list[RawCatalogItem]:
        url = "/v2/sizes"
        response = self._make_request("GET", url)
        return convert_response_to_raw_catalog_items(response)

    def _make_request(self, method: str, url: str):
        full_url = f"{self.api_url}{url}"
        params = {"per_page": 500}
        headers = {}
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"

        response = requests.request(
            method=method, url=full_url, params=params, headers=headers, timeout=30
        )
        response.raise_for_status()
        return response


def convert_response_to_raw_catalog_items(response) -> list[RawCatalogItem]:
    data = response.json()
    offers = []

    for size in data["sizes"]:
        gpu_info = size.get("gpu_info")
        if gpu_info:
            gpu_count = gpu_info["count"]
            gpu_vram_info = gpu_info["vram"]
            gpu_memory = float(gpu_vram_info["amount"])
            gpu_model = gpu_info["model"]
            # gpu_model uses patterns like "amd_mi300x", "nvidia_h100", "nvidia_rtx6000_ada"
            model_parts = gpu_model.split("_")
            gpu_name = "".join(part.upper() for part in model_parts[1:])
            gpu_vendor = get_gpu_vendor(gpu_name)
            if gpu_vendor is None:
                logger.warning(
                    f"Could not determine GPU vendor for model '{gpu_model}'. Skipping droplet '{size['slug']}'."
                )
                continue
        else:
            gpu_count = 0
            gpu_vendor = None
            gpu_name = ""
            gpu_memory = 0

        total_disk_size = sum(
            float(disk["size"]["amount"]) for disk in size["disk_info"] if disk["type"] == "local"
        )

        memory_gb = float(size["memory"]) / 1024  # MB -> GB

        # Creates an offer for each available region.
        # If regions list is empty, instance type is not available.
        for region in size["regions"]:
            offer = RawCatalogItem(
                instance_name=size["slug"],
                location=region,
                price=size["price_hourly"],
                cpu=size["vcpus"],
                memory=memory_gb,
                gpu_vendor=gpu_vendor,
                gpu_count=gpu_count,
                gpu_name=gpu_name,
                gpu_memory=gpu_memory,
                spot=False,
                disk_size=total_disk_size,
                flags=[],
            )
            offers.append(offer)

    return offers



import pytest

from gpuhunt import QueryFilter
from gpuhunt._internal.models import RawCatalogItem
from gpuhunt.providers.tensordock import TensorDockProvider


@pytest.fixture
def specs() -> dict:
    return {
        "cpu": {"amount": 256, "price": 0.003, "type": "Intel Xeon Platinum 8352Y"},
        "gpu": {
            "l40-pcie-48gb": {
                "amount": 8,
                "gtx": False,
                "pcie": True,
                "price": 1.05,
                "rtx": False,
                "vram": 48,
            }
        },
        "ram": {"amount": 1495, "price": 0.002},
        "storage": {"amount": 10252, "price": 5e-05},
    }


class TestTensorDockMinimalConfiguration:
    def test_no_requirements(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(QueryFilter(), specs, "", "")
        assert offers == make_offers(specs, cpu=16, memory=96, disk_size=96, gpu_count=1)

    def test_min_cpu_no_balance(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(
            QueryFilter(min_cpu=4), specs, "", "", balance_resources=False
        )
        assert offers == make_offers(specs, cpu=4, memory=96, disk_size=96, gpu_count=1)

    def test_min_cpu(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(QueryFilter(min_cpu=4), specs, "", "")
        assert offers == make_offers(specs, cpu=16, memory=96, disk_size=96, gpu_count=1)

    def test_too_many_min_cpu(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(QueryFilter(min_cpu=1000), specs, "", "")
        assert offers == []

    def test_min_memory_no_balance(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(
            QueryFilter(min_memory=3), specs, "", "", balance_resources=False
        )
        assert offers == make_offers(specs, cpu=2, memory=4, disk_size=48, gpu_count=1)

    def test_min_memory(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(QueryFilter(min_memory=3), specs, "", "")
        assert offers == make_offers(specs, cpu=16, memory=96, disk_size=96, gpu_count=1)

    def test_too_large_min_memory(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(QueryFilter(min_memory=2000), specs, "", "")
        assert offers == []

    def test_min_gpu_count(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(QueryFilter(min_gpu_count=2), specs, "", "")
        assert offers == make_offers(specs, cpu=32, memory=192, disk_size=192, gpu_count=2)

    def test_min_no_gpu(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(QueryFilter(max_gpu_count=0), specs, "", "")
        assert offers == []

    def test_min_total_gpu_memory(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(
            QueryFilter(min_total_gpu_memory=100), specs, "", ""
        )
        assert offers == make_offers(specs, cpu=48, memory=288, disk_size=288, gpu_count=3)

    def test_controversial_gpu(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(
            QueryFilter(min_total_gpu_memory=100, max_gpu_count=2), specs, "", ""
        )
        assert offers == []

    def test_all_cpu_all_gpu(self, specs: dict):
        offers = TensorDockProvider.optimize_offers(
            QueryFilter(min_cpu=256, min_gpu_count=1), specs, "", ""
        )
        assert offers == make_offers(specs, cpu=256, memory=768, disk_size=768, gpu_count=8)


def make_offers(
    specs: dict, cpu: int, memory: float, disk_size: float, gpu_count: int
) -> list[RawCatalogItem]:
    gpu = list(specs["gpu"].values())[0]
    price = cpu * specs["cpu"]["price"]
    price += memory * specs["ram"]["price"]
    price += disk_size * specs["storage"]["price"]
    price += gpu_count * gpu["price"]
    return [
        RawCatalogItem(
            instance_name="",
            location="",
            price=round(price, 5),
            cpu=cpu,
            memory=memory,
            gpu_vendor=None,
            gpu_count=gpu_count,
            gpu_name="L40",
            gpu_memory=gpu["vram"],
            spot=False,
            disk_size=disk_size,
        )
    ]



import json
import logging
import math
import os
import re
import time
from collections import namedtuple
from collections.abc import Iterable
from queue import Queue
from threading import Thread
from typing import Optional

import requests
import requests.adapters
from azure.core.credentials import TokenCredential
from azure.identity import DefaultAzureCredential
from azure.mgmt.compute import ComputeManagementClient

from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
prices_url = "https://prices.azure.com/api/retail/prices"
retail_prices_page_size = 1000
prices_version = "2023-01-01-preview"
prices_filters = [
    "serviceName eq 'Virtual Machines'",
    "priceType eq 'Consumption'",
    "contains(productName, 'Windows') eq false",
    "contains(productName, 'Dedicated') eq false",
    "contains(meterName, 'Low Priority') eq false",  # retires in 2025
]
VMSeries = namedtuple("VMSeries", ["pattern", "gpu_name", "gpu_memory"])
gpu_vm_series = [
    VMSeries(r"NC(\d+)ads_A100_v4", "A100", 80.0),  # NC A100 v4-series [A100 80GB]
    VMSeries(r"NC(\d+)ads_A10_v4", "A10", 24.0),  # NC A10 v4-series [A10]
    VMSeries(r"NC(\d+)as_T4_v3", "T4", 16.0),  # NCasT4_v3-series [T4]
    VMSeries(r"NC(\d+)r?s_v3", "V100", 16.0),  # NCv3-series [V100 16GB]
    VMSeries(r"ND(\d+)amsr_A100_v4", "A100", 80.0),  # NDm A100 v4-series [8xA100 80GB]
    VMSeries(r"ND(\d+)asr_v4", "A100", 40.0),  # ND A100 v4-series [8xA100 40GB]
    VMSeries(r"ND(\d+)rs_v2", "V100", 32.0),  # NDv2-series [8xV100 32GB]
    VMSeries(r"NG(\d+)adm?s_V620_v1", "V620", None),  # NGads V620-series [V620]  # todo
    VMSeries(r"NV(\d+)adm?s_A10_v5", "A10", 24.0),  # NVadsA10 v5-series [A10]
    VMSeries(r"NV(\d+)as_v4", "MI25", None),  # NVv4-series [MI25]  # todo
    VMSeries(r"NV(\d+)s_v3", "M60", None),  # NVv3-series [M60]  # todo
]
# https://learn.microsoft.com/en-us/azure/virtual-machines/sizes-previous-gen
retired_vm_series = [
    r"Basic_A(\d+)",
    r"Standard_A(\d+)",
    r"Standard_D(\d+)",
    r"Standard_DC(\d+)s",
    r"Standard_DS(\d+)",
    r"Standard_F(\d+)",
    r"Standard_F(\d+)s",
    r"Standard_G(\d+)",
    r"Standard_GS(\d+)",
    r"Standard_L(\d+)s",
    r"Standard_NC(\d+)r?",
    r"Standard_NC(\d+)r?s_v2",
    r"Standard_ND(\d+)r?s",
    r"Standard_NV(\d+)",
    r"Standard_NV(\d+)s_v2",
]


class AzureProvider(AbstractProvider):
    NAME = "azure"

    def __init__(
        self,
        subscription_id: str,
        credential: Optional[TokenCredential] = None,
        cache_dir: Optional[str] = None,
    ):
        self.cache_dir = cache_dir
        self.client = ComputeManagementClient(
            credential=credential or DefaultAzureCredential(),
            subscription_id=subscription_id,
        )

    def get_pages(self, threads: int = 8) -> Iterable[list[dict]]:
        q = Queue()
        workers = [
            Thread(target=self._get_pages_worker, args=(q, threads, i), daemon=True)
            for i in range(threads)
        ]
        for worker in workers:
            worker.start()

        exited = 0
        while exited < threads:
            page = q.get()
            if page is None:
                exited += 1
            else:
                yield page
            q.task_done()

    def _get_pages_worker(self, q: Queue, stride: int, worker_id: int):
        page_id = worker_id
        session = requests.Session()
        session.mount("https://", requests.adapters.HTTPAdapter(max_retries=3))
        try:
            while True:
                cached_page = None
                if self.cache_dir is not None:
                    cached_page = os.path.join(self.cache_dir, f"{page_id:04}.json")
                if cached_page is not None and os.path.exists(cached_page):
                    with open(cached_page) as f:
                        data = json.load(f)
                else:
                    logger.info("Worker %s fetches pricing page %s", worker_id, page_id)
                    res = session.get(
                        prices_url,
                        params={
                            "$filter": " and ".join(prices_filters),
                            "$skip": page_id * retail_prices_page_size,
                        },
                    )
                    if res.status_code == 429:
                        logger.warning("Worker %s got 429: sleep 3 & retry", worker_id)
                        time.sleep(3)
                        continue
                    res.raise_for_status()
                    if cached_page is not None:
                        with open(cached_page, "w") as f:
                            f.write(res.text)
                    data = res.json()
                if not data["Items"]:
                    logger.info("Worker %s exited", worker_id)
                    return
                q.put(data["Items"])
                page_id += stride
        except Exception as e:
            logger.exception("Worker %s failed: %s", worker_id, e)
        finally:
            q.put(None)

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        offers = []
        for page in self.get_pages():
            for item in page:
                if is_retired(item["armSkuName"]):
                    continue
                if not item["armSkuName"]:
                    continue
                price = float(item["retailPrice"])
                if math.isclose(price, 0):
                    continue
                offer = RawCatalogItem(
                    instance_name=item["armSkuName"],
                    location=item["armRegionName"],
                    price=price,
                    spot="Spot" in item["meterName"],
                    cpu=None,
                    memory=None,
                    gpu_vendor=None,
                    gpu_count=None,
                    gpu_name=None,
                    gpu_memory=None,
                    disk_size=None,
                )
                offers.append(offer)
        offers = self.fill_details(offers)
        return sorted(offers, key=lambda i: i.price)

    def fill_details(self, offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
        logger.info("Fetching instance details")
        instances = {}
        resources = self.client.resource_skus.list()
        for resource in resources:
            if resource.resource_type != "virtualMachines":
                continue
            if is_retired(resource.name):
                continue
            capabilities = {pair.name: pair.value for pair in resource.capabilities}
            cpu = capabilities.get("vCPUs")
            memory = capabilities.get("MemoryGB")
            if not cpu:
                logger.warning("Instance CPU is missing: %s", resource.name)
                continue
            if not memory:
                logger.warning("Instance memory is missing: %s", resource.name)
                continue
            gpu_count, gpu_name, gpu_memory = 0, None, None
            if "GPUs" in capabilities:
                gpu_count = int(capabilities["GPUs"])
                gpu_name, gpu_memory = get_gpu_name_memory(resource.name)
                if gpu_name is None and gpu_count:
                    logger.warning("Can't parse VM name: %s", resource.name)
                    continue
            instances[resource.name] = RawCatalogItem(
                instance_name=resource.name,
                cpu=cpu,
                memory=float(memory),
                gpu_vendor=None,
                gpu_count=gpu_count,
                gpu_name=gpu_name,
                gpu_memory=gpu_memory,
                location=None,
                price=None,
                spot=None,
                disk_size=None,
            )
        with_details = []
        for offer in offers:
            if (resources := instances.get(offer.instance_name)) is None:
                continue
            offer.cpu = resources.cpu
            offer.memory = resources.memory
            offer.gpu_count = resources.gpu_count
            offer.gpu_name = resources.gpu_name
            offer.gpu_memory = resources.gpu_memory
            offer.gpu_vendor = resources.gpu_vendor
            with_details.append(offer)
        return with_details

    @classmethod
    def filter(cls, offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
        vm_series = [
            VMSeries(r"D(\d+)s_v6", None, None),  # Dsv6-series
            VMSeries(
                r"E(2|4|8|16|20|32|48|64|96)s_v6", None, None
            ),  # Esv6-series (E128 and E192i are not yet GA)
            VMSeries(r"F(\d+)s_v2", None, None),  # Fsv2-series
            VMSeries(r"NC(\d+)s_v3", "V100", 16 * 1024),  # NCv3-series [V100 16GB]
            VMSeries(r"NC(\d+)as_T4_v3", "T4", 16 * 1024),  # NCasT4_v3-series [T4]
            VMSeries(r"ND(\d+)rs_v2", "V100", 32 * 1024),  # NDv2-series [8xV100 32GB]
            VMSeries(r"NV(\d+)adm?s_A10_v5", "A10", 24 * 1024),  # NVadsA10 v5-series [A10]
            VMSeries(r"NC(\d+)ads_A100_v4", "A100", 80 * 1024),  # NC A100 v4-series [A100 80GB]
            VMSeries(r"ND(\d+)asr_v4", "A100", 40 * 1024),  # ND A100 v4-series [8xA100 40GB]
            VMSeries(
                r"ND(\d+)amsr_A100_v4", "A100", 80 * 1024
            ),  # NDm A100 v4-series [8xA100 80GB]
            # The deprecated series are collected for older dstack versions
            VMSeries(
                r"D(\d+)s_v3", None, None
            ),  # Dsv3-series (deprecated in favor of Dsv6-series)
            VMSeries(
                r"E(\d+)i?s_v4", None, None
            ),  # Esv4-series (deprecated in favor of Esv6-series)
            VMSeries(
                r"E(\d+)-(\d+)s_v4", None, None
            ),  # Esv4-series (constrained vCPU, deprecated in favor of Esv6-series)
        ]
        vm_series_pattern = re.compile(
            f"^Standard_({'|'.join(series.pattern for series in vm_series)})$"
        )
        return [i for i in offers if vm_series_pattern.match(i.instance_name)]


def get_gpu_name_memory(vm_name: str) -> tuple[Optional[str], Optional[float]]:
    for pattern, gpu_name, gpu_memory in gpu_vm_series:
        m = re.match(f"^Standard_{pattern}$", vm_name)
        if m is None:
            continue
        if gpu_name == "A10" and vm_name.endswith("_v4"):
            gpu_memory = gpu_memory * min(1.0, int(m.group(1)) / 16)
        elif gpu_name == "A10" and vm_name.endswith("_v5"):
            gpu_memory = gpu_memory * min(1.0, int(m.group(1)) / 36)

        return gpu_name, gpu_memory
    return None, None


def is_retired(name: str) -> bool:
    if re.match(f"^({'|'.join(retired_vm_series)})$", name):
        return True
    return False



from pathlib import Path
from textwrap import dedent

import pytest

from gpuhunt.scripts.catalog_v1.__main__ import main


@pytest.mark.parametrize(
    ("v2_catalog", "v1_catalog"),
    [
        pytest.param(
            dedent(
                """
                instance_name,location,price,cpu,memory,gpu_count,gpu_name,gpu_memory,spot,disk_size,gpu_vendor,flags
                i1,us,0.5,30,240,1,A10,24.0,False,,nvidia,
                i2,us,1.0,30,240,2,A10,24.0,False,,nvidia,f1
                i3,us,2.0,30,240,4,A10,24.0,False,,nvidia,f1 f2
                i4,us,4.0,30,240,8,A10,24.0,False,,nvidia,
                """
            ).lstrip(),
            dedent(
                """
                instance_name,location,price,cpu,memory,gpu_count,gpu_name,gpu_memory,spot,disk_size,gpu_vendor
                i1,us,0.5,30,240,1,A10,24.0,False,,nvidia
                i4,us,4.0,30,240,8,A10,24.0,False,,nvidia
                """
            ).lstrip(),
            id="filters-out-offers-with-flags",
        ),
        pytest.param(
            dedent(
                """
                new_column,instance_name,location,price,cpu,memory,gpu_count,gpu_name,gpu_memory,spot,disk_size,gpu_vendor,flags
                ???,i1,us,0.5,30,240,1,A10,24.0,False,,nvidia,
                ???,i2,us,1.0,30,240,2,A10,24.0,False,,nvidia,
                """
            ).lstrip(),
            dedent(
                """
                instance_name,location,price,cpu,memory,gpu_count,gpu_name,gpu_memory,spot,disk_size,gpu_vendor
                i1,us,0.5,30,240,1,A10,24.0,False,,nvidia
                i2,us,1.0,30,240,2,A10,24.0,False,,nvidia
                """
            ).lstrip(),
            id="removes-extra-columns",
        ),
    ],
)
def test_main(tmp_path: Path, v2_catalog: str, v1_catalog: str) -> None:
    (tmp_path / "v1").mkdir()
    (tmp_path / "v2").mkdir()
    v1_file = tmp_path / "v1" / "catalog.csv"
    v2_file = tmp_path / "v2" / "catalog.csv"
    v2_file.write_text(v2_catalog)
    main(["--input", str(v2_file), "--output", str(v1_file)])
    assert v1_file.read_text() == v1_catalog



import logging
from math import ceil
from typing import Optional, TypeVar, Union

import requests

from gpuhunt._internal.constraints import get_compute_capability, is_between
from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)

# https://documenter.getpostman.com/view/20973002/2s8YzMYRDc#2b4a3db0-c162-438c-aae4-6a88afc96fdb
marketplace_hostnodes_url = "https://marketplace.tensordock.com/api/v0/client/deploy/hostnodes"
marketplace_gpus = {
    "a100-pcie-80gb": "A100",
    "geforcegtx1070-pcie-8gb": "GTX1070",
    "geforcertx3060-pcie-12gb": "RTX3060",
    "geforcertx3060ti-pcie-8gb": "RTX3060Ti",
    "geforcertx3060tilhr-pcie-8gb": "RTX3060TiLHR",
    "geforcertx3070-pcie-8gb": "RTX3070",
    "geforcertx3070ti-pcie-8gb": "RTX3070Ti",
    "geforcertx3080-pcie-10gb": "RTX3080",
    "geforcertx3080ti-pcie-12gb": "RTX3080Ti",
    "geforcertx3090-pcie-24gb": "RTX3090",
    "geforcertx4090-pcie-24gb": "RTX4090",
    "l40-pcie-48gb": "L40",
    "rtxa4000-pcie-16gb": "A4000",
    "rtxa5000-pcie-24gb": "A5000",
    "rtxa6000-pcie-48gb": "A6000",
    "v100-pcie-16gb": "V100",
}

RAM_PER_VRAM = 2
RAM_PER_CORE = 6
CPU_DIV = 2  # has to be even
RAM_DIV = 2  # has to be even


class TensorDockProvider(AbstractProvider):
    NAME = "tensordock"

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        logger.info("Fetching TensorDock offers")

        hostnodes = requests.get(marketplace_hostnodes_url, timeout=10).json()["hostnodes"]
        offers = []
        for hostnode, details in hostnodes.items():
            location = details["location"]["country"].lower().replace(" ", "")
            if query_filter is not None:
                offers += self.optimize_offers(
                    query_filter,
                    details["specs"],
                    hostnode,
                    location,
                    balance_resources=balance_resources,
                )
            else:  # pick maximum possible configuration
                for gpu_name, gpu in details["specs"]["gpu"].items():
                    if gpu["amount"] == 0:
                        continue
                    offers.append(
                        RawCatalogItem(
                            instance_name=hostnode,
                            location=location,
                            price=round(
                                sum(
                                    details["specs"][key]["price"]
                                    * details["specs"][key]["amount"]
                                    for key in ("cpu", "ram", "storage")
                                )
                                + gpu["amount"] * gpu["price"],
                                5,
                            ),
                            cpu=round_down(details["specs"]["cpu"]["amount"], 2),
                            memory=float(round_down(details["specs"]["ram"]["amount"], 2)),
                            gpu_vendor=None,
                            gpu_count=gpu["amount"],
                            gpu_name=convert_gpu_name(gpu_name),
                            gpu_memory=float(gpu["vram"]),
                            spot=False,
                            disk_size=float(details["specs"]["storage"]["amount"]),
                        )
                    )
        return sorted(offers, key=lambda i: i.price)

    @staticmethod
    def optimize_offers(
        q: QueryFilter,
        specs: dict,
        instance_name: str,
        location: str,
        balance_resources: bool = True,
    ) -> list[RawCatalogItem]:
        """
        Picks the best offer for the given query filter
        Doesn't respect max values, additional filtering is required

        Args:
            q: query filter
            specs: hostnode specs
            instance_name: hostnode `instance_name`
            location: hostnode `location`
            balance_resources: if True, will override query filter min values
        """
        offers = []
        for gpu_model, gpu_info in specs["gpu"].items():
            # filter by single gpu characteristics
            if not is_between(gpu_info["vram"], q.min_gpu_memory, q.max_gpu_memory):
                continue
            gpu_name = convert_gpu_name(gpu_model)
            if q.gpu_name is not None and gpu_name.lower() not in map(str.lower, q.gpu_name):
                continue
            cc = get_compute_capability(gpu_name)
            if not cc or not is_between(cc, q.min_compute_capability, q.max_compute_capability):
                continue

            for gpu_count in range(1, gpu_info["amount"] + 1):  # try all possible gpu counts
                if not is_between(gpu_count, q.min_gpu_count, q.max_gpu_count):
                    continue
                total_gpu_memory = gpu_count * gpu_info["vram"]
                if not is_between(
                    total_gpu_memory, q.min_total_gpu_memory, q.max_total_gpu_memory
                ):
                    continue

                # we can't take 100% of CPU/RAM/storage if we don't take all GPUs
                multiplier = 0.75 if gpu_count < gpu_info["amount"] else 1
                available_memory = round_down(multiplier * specs["ram"]["amount"], RAM_DIV)
                available_cpu = round_down(multiplier * specs["cpu"]["amount"], CPU_DIV)
                available_disk = int(multiplier * specs["storage"]["amount"])

                memory = None
                if q.min_memory is not None:
                    if q.min_memory > available_memory:
                        continue
                    memory = round_up(
                        max_none(
                            q.min_memory,
                            gpu_count,  # 1 GB per GPU at least
                            q.min_cpu,  # 1 GB per CPU at least
                        ),
                        RAM_DIV,
                    )
                if memory is None or balance_resources:
                    memory = max_none(
                        memory,
                        min_none(
                            available_memory,
                            round_up(RAM_PER_VRAM * total_gpu_memory, RAM_DIV),
                            round_down(q.max_memory, RAM_DIV),  # can be None
                        ),
                    )

                cpu = None
                if q.min_cpu is not None:
                    if q.min_cpu > available_cpu:
                        continue
                    # 1 CPU per GPU at least
                    cpu = round_up(max(q.min_cpu, gpu_count), CPU_DIV)
                if cpu is None or balance_resources:
                    cpu = max_none(
                        cpu,
                        min_none(
                            available_cpu,
                            round_up(ceil(memory / RAM_PER_CORE), CPU_DIV),
                            round_down(q.max_cpu, CPU_DIV),  # can be None
                        ),
                    )

                disk_size = None
                if q.min_disk_size is not None:
                    if q.min_disk_size > available_disk:
                        continue
                    disk_size = q.min_disk_size
                if disk_size is None or balance_resources:
                    disk_size = max_none(
                        disk_size,
                        min_none(
                            available_disk,
                            max(memory, total_gpu_memory),
                            q.max_disk_size,  # can be None
                        ),
                    )

                price = round(
                    memory * specs["ram"]["price"]
                    + cpu * specs["cpu"]["price"]
                    + disk_size * specs["storage"]["price"]
                    + gpu_count * gpu_info["price"],
                    5,
                )

                offer = RawCatalogItem(
                    instance_name=instance_name,
                    location=location,
                    price=price,
                    cpu=cpu,
                    memory=float(memory),
                    gpu_vendor=None,
                    gpu_name=gpu_name,
                    gpu_count=gpu_count,
                    gpu_memory=float(gpu_info["vram"]),
                    spot=False,
                    disk_size=disk_size,
                )
                offers.append(offer)
                break  # stop increasing gpu count
        return offers


def round_up(value: Optional[Union[int, float]], step: int) -> Optional[int]:
    if value is None:
        return None
    return round_down(value + step - 1, step)


def round_down(value: Optional[Union[int, float]], step: int) -> Optional[int]:
    if value is None:
        return None
    return value // step * step


T = TypeVar("T", bound=Union[int, float])


def min_none(*args: Optional[T]) -> T:
    return min(v for v in args if v is not None)


def max_none(*args: Optional[T]) -> T:
    return max(v for v in args if v is not None)


def convert_gpu_name(model: str) -> str:
    """
    >>> convert_gpu_name("geforcegtx1070-pcie-8gb")
    'GTX1070'
    >>> convert_gpu_name("geforcertx1111ti-pcie-13gb")
    'RTX1111Ti'
    >>> convert_gpu_name("a100-pcie-40gb")
    'A100'
    """
    if model in marketplace_gpus:
        return marketplace_gpus[model]
    model = model.split("-")[0]
    prefix = "geforce"
    if model.startswith(prefix):
        model = model[len(prefix) :]
    return model.upper().replace("TI", "Ti")



import copy
import logging
import re
from collections.abc import Iterable
from dataclasses import asdict, dataclass
from typing import Annotated, Optional

import oci
from oci.identity.models import Region
from pydantic import BaseModel, Field
from requests import Session
from typing_extensions import TypedDict

from gpuhunt._internal.constraints import find_accelerators
from gpuhunt._internal.models import AcceleratorVendor, QueryFilter, RawCatalogItem
from gpuhunt._internal.utils import to_camel_case
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
COST_ESTIMATOR_URL_TEMPLATE = "https://www.oracle.com/a/ocom/docs/cloudestimator2/data/{resource}"
COST_ESTIMATOR_REQUEST_TIMEOUT = 10


class OCICredentials(TypedDict):
    user: Optional[str]
    key_content: Optional[str]
    fingerprint: Optional[str]
    tenancy: Optional[str]
    region: Optional[str]


class OCIProvider(AbstractProvider):
    NAME = "oci"

    def __init__(self, credentials: OCICredentials):
        self.api_client = oci.identity.IdentityClient(
            credentials if all(credentials.values()) else oci.config.from_file()
        )
        self.cost_estimator = CostEstimator()

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        shapes = self.cost_estimator.get_shapes()
        products = self.cost_estimator.get_products()
        regions: list[Region] = self.api_client.list_regions().data

        result = []

        for shape in shapes.items:
            if (
                shape.hidden
                or shape.status != "ACTIVE"
                or shape.shape_type.value not in ("vm", "bm")
                or shape.sub_type.value not in ("standard", "gpu", "optimized")
                or ".A1." in shape.name
            ):
                continue

            try:
                resources = shape_to_resources(shape, products)
            except CostEstimatorDataError as e:
                logger.warning(
                    "Skipping shape %s due to unexpected Cost Estimator data: %s", shape.name, e
                )
                continue

            on_demand_item = RawCatalogItem(
                instance_name=shape.name,
                location=None,
                price=resources.total_price(),
                cpu=resources.cpu.vcpus,
                memory=resources.memory.gbs,
                gpu_vendor=None,
                gpu_count=resources.gpu.units_count,
                gpu_name=resources.gpu.name,
                gpu_memory=resources.gpu.unit_memory_gb,
                spot=False,
                disk_size=None,
            )
            item_variations = [on_demand_item]
            if shape.allow_preemptible:
                item_variations.append(self._make_spot_item(on_demand_item))
            for item in item_variations:
                result.extend(self._duplicate_item_in_regions(item, regions))

        return sorted(result, key=lambda i: i.price)

    @staticmethod
    def _make_spot_item(item: RawCatalogItem) -> RawCatalogItem:
        item = copy.deepcopy(item)
        item.spot = True
        # > Preemptible capacity costs 50% less than on-demand capacity
        # https://docs.oracle.com/en-us/iaas/Content/Compute/Concepts/preemptible.htm#howitworks__billing
        item.price *= 0.5
        item.flags.append("oci-spot")
        return item

    @staticmethod
    def _duplicate_item_in_regions(
        item: RawCatalogItem, regions: Iterable[Region]
    ) -> list[RawCatalogItem]:
        result = []
        for region in regions:
            regional_item = copy.deepcopy(item)
            regional_item.location = region.name
            result.append(regional_item)
        return result


class CostEstimatorTypeField(BaseModel):
    value: str


class CostEstimatorShapeProduct(BaseModel):
    type: CostEstimatorTypeField
    part_number: str
    qty: Optional[int]

    class Config:
        alias_generator = to_camel_case


class CostEstimatorShape(BaseModel):
    name: str
    hidden: bool
    status: str
    allow_preemptible: bool
    bundle_memory_qty: Optional[int]
    gpu_qty: Optional[int]
    gpu_memory_qty: Optional[int]
    processor_type: CostEstimatorTypeField
    shape_type: CostEstimatorTypeField
    sub_type: CostEstimatorTypeField
    products: list[CostEstimatorShapeProduct]

    class Config:
        alias_generator = to_camel_case

    def is_arm_cpu(self):
        is_ampere_gpu = self.sub_type.value == "gpu" and (
            "GPU4" in self.name or "GPU.A10" in self.name
        )
        # the data says A10 and A100 GPU instances are ARM, but they are not
        return self.processor_type.value == "arm" and not is_ampere_gpu

    def get_gpu_unit_memory_gb(self) -> Optional[float]:
        if self.gpu_memory_qty and self.gpu_qty:
            return self.gpu_memory_qty / self.gpu_qty
        return None


class CostEstimatorShapeList(BaseModel):
    items: list[CostEstimatorShape]


class CostEstimatorPrice(BaseModel):
    model: str
    value: float


class CostEstimatorPriceLocalization(BaseModel):
    currency_code: str
    prices: list[CostEstimatorPrice]

    class Config:
        alias_generator = to_camel_case


class CostEstimatorProduct(BaseModel):
    part_number: str
    billing_model: str
    price_type: Annotated[str, Field(alias="pricetype")]
    currency_code_localizations: list[CostEstimatorPriceLocalization]

    class Config:
        alias_generator = to_camel_case

    def find_price_l10n(self, currency_code: str) -> Optional[CostEstimatorPriceLocalization]:
        return next(
            filter(
                lambda price: price.currency_code == currency_code,
                self.currency_code_localizations,
            ),
            None,
        )


class CostEstimatorProductList(BaseModel):
    items: list[CostEstimatorProduct]

    def find(self, part_number: str) -> Optional[CostEstimatorProduct]:
        return next(filter(lambda product: product.part_number == part_number, self.items), None)


class CostEstimator:
    def __init__(self):
        self.session = Session()

    def get_shapes(self) -> CostEstimatorShapeList:
        return self._get("shapes.json", CostEstimatorShapeList)

    def get_products(self) -> CostEstimatorProductList:
        return self._get("products.json", CostEstimatorProductList)

    def _get(self, resource: str, ResponseModel: type[BaseModel]):
        url = COST_ESTIMATOR_URL_TEMPLATE.format(resource=resource)
        resp = self.session.get(url, timeout=COST_ESTIMATOR_REQUEST_TIMEOUT)
        resp.raise_for_status()
        return ResponseModel.parse_raw(resp.content)


class CostEstimatorDataError(Exception):
    pass


@dataclass
class CPUConfiguration:
    vcpus: int
    price: float


@dataclass
class MemoryConfiguration:
    gbs: int
    price: float


@dataclass
class GPUConfiguration:
    units_count: int
    unit_memory_gb: Optional[float]
    name: Optional[str]
    price: float

    def __post_init__(self):
        d = asdict(self)
        if any(d.values()) and not all(d.values()):
            raise CostEstimatorDataError(f"Incomplete GPU parameters: {self}")


@dataclass
class ResourcesConfiguration:
    cpu: CPUConfiguration
    memory: MemoryConfiguration
    gpu: GPUConfiguration

    def total_price(self) -> float:
        return self.cpu.price + self.memory.price + self.gpu.price


def shape_to_resources(
    shape: CostEstimatorShape, products: CostEstimatorProductList
) -> ResourcesConfiguration:
    cpu = None
    gpu = GPUConfiguration(units_count=0, unit_memory_gb=None, name=None, price=0.0)
    memory: Optional[MemoryConfiguration] = None
    if shape.bundle_memory_qty is not None:
        memory = MemoryConfiguration(gbs=shape.bundle_memory_qty, price=0.0)

    for product in shape.products:
        if product.qty is None:
            raise CostEstimatorDataError("Product quantity not found")
        product_details = products.find(product.part_number)
        if product_details is None:
            raise CostEstimatorDataError(f"Could not find product {product.part_number!r}")
        product_price = get_product_price_usd_per_hour(product_details)

        if product.type.value == "ocpu":
            vcpus = product.qty if shape.is_arm_cpu() else product.qty * 2
            if shape.gpu_qty:
                gpu = GPUConfiguration(
                    units_count=shape.gpu_qty,
                    unit_memory_gb=shape.get_gpu_unit_memory_gb(),
                    name=get_gpu_name(shape.name),
                    price=product_price * shape.gpu_qty,
                )
                cpu = CPUConfiguration(vcpus=vcpus, price=0.0)
            else:
                cpu = CPUConfiguration(vcpus=vcpus, price=product_price * product.qty)

        elif product.type.value == "memory":
            memory = MemoryConfiguration(gbs=product.qty, price=product_price * product.qty)

        else:
            raise CostEstimatorDataError(f"Unknown product type {product.type.value!r}")

    if cpu is None:
        raise CostEstimatorDataError("No ocpu product")
    if memory is None:
        raise CostEstimatorDataError("No memory product")

    return ResourcesConfiguration(cpu, memory, gpu)


def get_product_price_usd_per_hour(product: CostEstimatorProduct) -> float:
    if product.billing_model != "UCM":
        raise CostEstimatorDataError(
            f"Billing model for product {product.part_number!r} is {product.billing_model!r}"
        )
    if product.price_type != "HOUR":
        raise CostEstimatorDataError(
            f"Price type for product {product.part_number!r} is {product.price_type!r}"
        )
    price_l10n = product.find_price_l10n("USD")
    if price_l10n is None:
        raise CostEstimatorDataError(f"No USD price for product {product.part_number!r}")
    if len(price_l10n.prices) != 1:
        raise CostEstimatorDataError(
            f"Product {product.part_number!r} has {len(price_l10n.prices)} USD prices"
        )
    price = price_l10n.prices[0]
    if price.model != "PAY_AS_YOU_GO":
        raise CostEstimatorDataError(
            f"Pricing model for product {product.part_number!r} is {price.model!r}"
        )
    return price.value


def get_gpu_name(shape_name: str) -> Optional[str]:
    parts = re.split(r"[\.-]", shape_name.upper())

    if "GPU4" in parts:
        return "A100"
    if "GPU3" in parts:
        return "V100"
    if "GPU2" in parts:
        return "P100"

    if "GPU" in parts:
        gpu_name_index = parts.index("GPU") + 1
        if gpu_name_index < len(parts):
            gpu_name = parts[gpu_name_index]

            if accelerators := find_accelerators(
                names=[gpu_name], vendors=[AcceleratorVendor.NVIDIA]
            ):
                return accelerators[0].name
    return None



import csv
from collections import Counter
from pathlib import Path

import pytest

from gpuhunt.providers.runpod import GPU_MAP


@pytest.fixture
def data_rows(catalog_dir: Path) -> list[dict]:
    file = catalog_dir / "runpod.csv"
    reader = csv.DictReader(file.open())
    return list(reader)


def select_row(rows, name: str) -> list[str]:
    return [r[name] for r in rows if r[name]]


def test_locations(data_rows):
    expected = {
        # Secure cloud
        "CA-MTL-1",
        "CA-MTL-2",
        "CA-MTL-3",
        "EU-NL-1",
        "EU-RO-1",
        "EU-SE-1",
        "EUR-IS-1",
        "EUR-IS-2",
        "US-TX-3",
        # Community cloud
        "CA",
        "CZ",
        "FR",
        "US",
    }
    locations = set(select_row(data_rows, "location"))
    # Assert most are present. Some may be missing due to low availability
    assert len(expected - locations) <= 3


def test_spot(data_rows):
    spots = select_row(data_rows, "spot")

    expected = set(("True", "False"))
    assert set(spots) == expected

    count = Counter(spots)
    for spot_key in ("True", "False"):
        assert count[spot_key] > 1


def test_gpu_present(data_rows):
    refs = set(name for _, name in GPU_MAP.values())
    gpus = set(select_row(data_rows, "gpu_name"))
    assert len(refs & gpus) > 7
